'''
XbrlSemanticGraphDB.py implements a graph database interface for Arelle, based
on a concrete realization of the Abstract Model PWD 2.0 layer.  This is a semantic
representation of XBRL information.

This module provides the execution context for saving a dts and instances in
XBRL Rexter-interfaced graph.  It may be loaded by Arelle's RSS feed, or by individual
DTS and instances opened by interactive or command line/web service mode.

See COPYRIGHT.md for copyright information.

to do:

1) add AMTF cube regions (dimensions)
    HF - don't believe this is either feasible or has a use case in a graph model
2) check existence of (shared) documents and contained elements before adding
3) tuple structure declaration (particles in elements of data dictionary?)
4) tuple structure (instance facts)
5) add footnote resources to relationships (and test with EDInet footnote references)
6) test some filings with text blocks (shred them?)  (30mB - 50mB sized text blocks?)
7) add mappings to, or any missing relationships, of Charlie's financial model

'''

import os, io, time, json, socket, logging, zlib
from math import isnan, isinf
from arelle.ModelDtsObject import ModelConcept, ModelResource, ModelRelationship
from arelle.ModelInstanceObject import ModelFact, ModelInlineFact
from arelle.ModelDocument import Type
from arelle.ModelValue import qname, datetime
from arelle.ValidateXbrlCalcs import roundValue
from arelle import XbrlConst, XmlUtil
import urllib.request
from urllib.error import HTTPError, URLError

TRACEGREMLINFILE = None
#TRACEGREMLINFILE = r"c:\temp\rexstertrace.log"  # uncomment to trace SQL on connection (very big file!!!)

def insertIntoDB(modelXbrl,
                 user=None, password=None, host=None, port=None, database=None, timeout=None,
                 product=None, rssItem=None, **kwargs):
    db = None
    try:
        xsgdb = XbrlSemanticGraphDatabaseConnection(modelXbrl, user, password, host, port, database, timeout)
        xsgdb.verifyGraphs()
        xsgdb.insertXbrl(rssItem=rssItem)
        xsgdb.close()
    except Exception as ex:
        if xsgdb is not None:
            try:
                xsgdb.close(rollback=True)
            except Exception as ex2:
                pass
        raise # reraise original exception with original traceback

def isDBPort(host, port, timeout=10):
    # determine if postgres port
    t = 2
    while t < timeout:
        try:
            conn = urllib.request.urlopen("http://{0}:{1}/graphs".format(host, port or '8182'))
            return True # success but doesn't need password
        except HTTPError as err:
            return False # success, this is really a postgres socket, wants user name
        except URLError:
            return False # something is there but not postgres
        except socket.timeout:
            t = t + 2  # relax - try again with longer timeout
    return False

XBRLDBGRAPHS = {
                "filings",       # filings (graph of reports)
                "documents"      # graph of namespace->names->types/elts, datapoints
                                 # any future root vertices go here
                }

HTTPHEADERS = {'User-agent':   'Arelle/1.0',
               'Accept':       'application/json',
               'Content-Type': 'application/json'}

def pyBoolFromDbBool(str):
    return str in ("TRUE", "t")

def pyNoneFromDbNULL(str):
    return None

def dbNum(num):
    if isinstance(num, (int,float)):
        if isinf(num) or isnan(num):
            return None  # not legal in SQL
        return num
    return None

def dbString(s): # compress long strings
    if isinstance(s, str) and len(s) > 512:
        return ''.join(map(chr,zlib.compress(s.encode()))) # compress as utf-8 but return as string
    return s

class XPDBException(Exception):
    def __init__(self, code, message, **kwargs ):
        self.code = code
        self.message = message
        self.kwargs = kwargs
        self.args = ( self.__repr__(), )
    def __repr__(self):
        return _('[{0}] exception: {1}').format(self.code, self.message % self.kwargs)



class XbrlSemanticGraphDatabaseConnection():
    def __init__(self, modelXbrl, user, password, host, port, database, timeout, product):
        self.modelXbrl = modelXbrl
        self.disclosureSystem = modelXbrl.modelManager.disclosureSystem
        #self.conn = RexProConnection(host, int(port or '8182'), (database or 'emptygraph'),
        #                             user=user, password=password)
        connectionUrl = "http://{0}:{1}".format(host, port or '8182')
        self.url = connectionUrl + '/graphs/' + database
        # Create an OpenerDirector with support for Basic HTTP Authentication...
        auth_handler = urllib.request.HTTPBasicAuthHandler()
        if user:
            auth_handler.add_password(realm='rexster',
                                      uri=connectionUrl,
                                      user=user,
                                      passwd=password)
        self.conn = urllib.request.build_opener(auth_handler)
        self.timeout = timeout or 60
        self.verticePropTypes = {}

    def close(self, rollback=False):
        try:
            self.conn.close()
            self.__dict__.clear() # dereference everything
        except Exception as ex:
            self.__dict__.clear() # dereference everything
            raise

    @property
    def isClosed(self):
        return not bool(self.__dict__)  # closed when dict is empty

    def showStatus(self, msg, clearAfter=None):
        self.modelXbrl.modelManager.showStatus(msg, clearAfter)

    def verifyGraphs(self):
        # if no tables, initialize database
        missingRoots = XBRLDBGRAPHS - self.loadGraphRootVertices()
        if missingRoots:  # some are missing
            raise XPDBException("xsgDB:MissingGraphs",
                                _("The following graph roots are missing: %(missingRootNames)s"),
                                missingRootNames=', '.join(t for t in sorted(missingRoots)))

    def execute(self, activity, script, params=None, commit=False, close=True, fetch=True):
        gremlin = {"script": script}
        if params:
            gremlin["params"] = params
        if TRACEGREMLINFILE:
            with io.open(TRACEGREMLINFILE, "a", encoding='utf-8') as fh:
                fh.write("\n\n>>> sent: \n{0}".format(str(gremlin)))
        request = urllib.request.Request(self.url + "/tp/gremlin",
                                         data=json.dumps(gremlin, ensure_ascii=False).encode('utf-8'),
                                         headers=HTTPHEADERS)
        try:
            with self.conn.open(request, timeout=self.timeout) as fp:
                results = json.loads(fp.read().decode('utf-8'))
        except HTTPError as err:
            if err.code == 500: # results are not successful but returned nontheless
                results = json.loads(err.fp.read().decode('utf-8'))
            else:
                raise  # reraise any other errors
        if TRACEGREMLINFILE:
            with io.open(TRACEGREMLINFILE, "a", encoding='utf-8') as fh:
                fh.write("\n\n>>> received: \n{0}".format(str(results)))
        if results.get('success', False) == False:
            raise XPDBException("xsgDB:DatabaseError",
                                _("%(activity)s not successful: %(error)s"),
                                activity=activity, error=results.get('error'))
        return results

    def commit(self):
        self.execute("Commit transaction", "g.commit()")

    def rollback(self):
        self.execute("Rollback transaction", "g.rollback()")

    def loadGraphRootVertices(self):
        self.showStatus("Load/Create graph root vertices")
        # try to create root index
        results = self.execute("Load/Create graph root vertices", """
            def r, v
            // vertex index
            try {  // not all gremlin servers support key index
                if (!("_rlkey" in g.getIndexedKeys(Vertex.class))) {
                    g.createKeyIndex("_rlkey", Vertex.class)
                    //g.createKeyIndex("_class", Vertex.class)
                    g.commit()
                }
            } catch (Exception e) {
            }
            // check if semantic_root vertex already exists
            rIt = g.V('_rlkey', 'semantic_root')  // iterator on semantic_root vertices
            // if none, add it
            if (rIt.hasNext()) {
                r = rIt.next()
            } else {
                r = g.addVertex(['_class':'semantic_root', '_rlkey':'semantic_root'])
            }
            root_vertices = []
            root_classes.each{
                // check if class "it"'s vertex already exists
                vIt = r.out(it)
                // if exists, use that vertex, if not, add it
                if (vIt.hasNext()) {
                    v = vIt.next()
                } else {
                    v = g.addVertex(['_class':it])
                    g.addEdge(r, v, it)
                }
                // return vertex (so plug-in can refer to it by its id
                root_vertices << v
            }
            root_vertices
            """,
            params={"root_classes":list(XBRLDBGRAPHS)})["results"]
        for v in results:
            setattr(self, "root_" + v['_class'] + "_id", int(v['_id']))
        return set(v['_class'] for v in results)

    def getDBsize(self):
        self.showStatus("Get database size (slow operation for now)")
        results = self.execute("Get database size", """
            [g.V.count(), g.E.count()]
            """)["results"]
        return (results[0], results[1])

    def insertXbrl(self, rssItem):
        try:
            # must also have default dimensions loaded
            from arelle import ValidateXbrlDimensions
            ValidateXbrlDimensions.loadDimensionDefaults(self.modelXbrl)

            #initialVcount, initialEcount = self.getDBsize() # don't include in timing, very slow
            startedAt = time.time()
            # self.load()  this done in the verify step
            self.insertFiling(rssItem)
            self.insertDocuments()
            self.insertDataDictionary() # XML namespaces types aspects
            #self.insertRelationshipTypeSets()
            #self.insertResourceRoleSets()
            #self.insertAspectValues()
            self.modelXbrl.profileStat(_("XbrlPublicDB: geport insertion"), time.time() - startedAt)
            startedAt = time.time()
            self.insertDataPoints()
            self.modelXbrl.profileStat(_("XbrlPublicDB: data points insertion"), time.time() - startedAt)
            startedAt = time.time()
            self.insertRelationshipSets()
            self.modelXbrl.profileStat(_("XbrlPublicDB: Relationships insertion"), time.time() - startedAt)
            self.insertValidationResults()
            self.modelXbrl.profileStat(_("XbrlPublicDB: Validation results insertion"), time.time() - startedAt)
            #startedAt = time.time()
            #self.insertValidCombinations()
            #self.modelXbrl.profileStat(_("XbrlPublicDB: Valid Combinations insertion"), time.time() - startedAt)
            self.showStatus("Committing entries")
            self.commit()
            self.modelXbrl.profileStat(_("XbrlPublicDB: insertion committed"), time.time() - startedAt)
            #finalVcount, finalEcount = self.getDBsize()
            #self.modelXbrl.modelManager.addToLog("added vertices: {0}, edges: {1}, total vertices: {2}, edges: {3}".format(
            #              finalVcount - initialVcount, finalEcount - initialEcount, finalVcount, finalEcount))
            self.showStatus("DB insertion completed", clearAfter=5000)
        except Exception as ex:
            self.showStatus("DB insertion failed due to exception", clearAfter=5000)
            self.rollback()
            raise

    def insertFiling(self, rssItem):
        self.showStatus("insert filing")
        # filing graph -> document vertices
        new_filing = {'_class':'filing',
                         'is_most_current': True}
        if self.modelXbrl.modelDocument.creationSoftwareComment:
            new_filing['creation_software'] = self.modelXbrl.modelDocument.creationSoftwareComment
        datetimeNow = datetime.datetime.now()
        datetimeNowStr = XmlUtil.dateunionValue(datetimeNow)
        if rssItem is not None:  # sec filing (accession)
            # set self.
            filingType = "SEC_filing"
            # for an RSS Feed entry from SEC, use rss item's filing information
            new_filing['accepted_timestamp'] = XmlUtil.dateunionValue(rssItem.acceptanceDatetime)
            new_filing['filing_date'] = XmlUtil.dateunionValue(rssItem.filingDate)
            new_filing['entity_id'] = rssItem.cikNumber
            new_filing['entity_name'] = rssItem.companyName
            new_filing['standard_industrial_classification'] = rssItem.assignedSic
            new_filing['sec_html_url'] = rssItem.htmlUrl
            new_filing['entry_url'] = rssItem.url
            new_filing['filing_number'] = filing_number = rssItem.accessionNumber
        else:
            # not an RSS Feed item, make up our own filing ID (the time in seconds of epoch)
            intNow = int(time.time())
            filingType = "independent_filing"
            new_filing['accepted_timestamp'] = datetimeNowStr
            new_filing['filing_date'] = datetimeNowStr
            new_filing['entry_url'] = self.modelXbrl.fileSource.url
            new_filing['filing_number'] = filing_number = str(intNow)
        for id in self.execute("Insert filing " + filingType, """
            r = g.v(root_filings_id)
            // check if filing already has a vertex
            vIt = r.out(new_filing.filing_number)
            // use prior vertex, or if none, create new vertex for it
            filing = (vIt.hasNext() ? vIt.next() : g.addVertex(new_filing) )
            // TBD: modify filing timestamp (last-updated-at, if it already existed)
            // check if vertex has edge to root_filings vertex
            vIn = filing.in
            // if no edge, add one
            vIn.hasNext() && vIn.next() == r ?: g.addEdge(r, filing, new_filing.filing_number)
            filing.id
            """,
            params={'root_filings_id': self.root_filings_id,
                    'new_filing': new_filing,
                    'filing_type': filingType,
                    'datetime_now': datetimeNowStr,
                   })["results"]:
            self.filing_id = int(id)

        # relationshipSets are a dts property
        self.relationshipSets = [(arcrole, ELR, linkqname, arcqname)
                                 for arcrole, ELR, linkqname, arcqname in self.modelXbrl.baseSets.keys()
                                 if ELR and (arcrole.startswith("XBRL-") or (linkqname and arcqname))]

    def insertDocuments(self):
        # filing->documents
        #
        self.showStatus("insert documents")
        documents = []
        for modelDocument in self.modelXbrl.urlDocs.values():
            doc = {'_class': 'document',
                   'url': modelDocument.uri,
                   'document_type': modelDocument.gettype()}
            documents.append(doc)
        results = self.execute("Insert documents", """
            results = []
            rDoc = g.v(root_documents_id)
            vFiling = g.v(filing_id)
            // add report if it doesn't exist
            vReportIt = vFiling.out('reports')
            vReport = (vReportIt.hasNext() ? vReportIt.next() : g.addVertex(report) )
            vReportIn = vReport.in('reports').has('id',vFiling.id)
            vReportIn.hasNext() ?: g.addEdge(vFiling, vReport, 'reports')

            urlV = [:]
            urlV_id = [:]
            isNew = [:]
            documents.each{
                vDocIt = rDoc.out(it.url)
                isNew[it.url] = !vDocIt.hasNext()
                vDoc = (vDocIt.hasNext() ? vDocIt.next() : g.addVertex(it))
                // link doc to root doc
                vDocIn = vDoc.in(it.url)
                vDocIn.hasNext() && vDocIn.next() == rDoc ?: g.addEdge(rDoc, vDoc, it.url)
                urlV[it.url] = vDoc
                urlV_id[it.url] = vDoc.id
            }

            // entry document edge to doc root and report
            documents.findAll{it.url == entry_url}.each{
                vEntryDoc = urlV[it.url]
                // link entry point entryDoc to report
                vDocIn = vReport.in('entry_point')
                vDocIn.hasNext() && vDocIn.next() == vReport ?: g.addEdge(vReport, vEntryDoc, 'entry_point')
                vDocIn = vEntryDoc.in('filed_document')
                vDocIn.hasNext() && vDocIn.next() == rDoc ?: g.addEdge(vReport, vEntryDoc, 'filed_document')
            }

            // referenced document edge to entry document
            documents.findAll{it.url != entry_url}.each{
                vRefDoc = urlV[it.url]
                // link refDoc to vEntryDoc
                vDocIn = vRefDoc.in('referenced_document')
                vDocIn.hasNext() && vDocIn.next() == vEntryDoc ?: g.addEdge(vEntryDoc, vRefDoc, 'referenced_document')
            }
            [vReport.id, urlV_id, isNew]
            """,
            params={'root_documents_id': self.root_documents_id,
                    'filing_id': self.filing_id,
                    'entry_url': self.modelXbrl.modelDocument.uri,
                    'report': {
                        '_class': 'report'},
                    'documents': documents
                    })["results"]
        report_id, doc_id_list, doc_isNew_list = results # unpack list
        self.report_id = int(report_id)
        self.document_ids = dict( (url, int(id)) for url, id in doc_id_list.items() )
        self.document_isNew = dict( (url, isNew) for url,isNew in doc_isNew_list.items() )

    def conceptsUsed(self):
        conceptsUsed = set(f.qname for f in self.modelXbrl.factsInInstance)

        for cntx in self.modelXbrl.contexts.values():
            for dim in cntx.qnameDims.values():
                conceptsUsed.add(dim.dimensionQname)
                if dim.isExplicit:
                    conceptsUsed.add(dim.memberQname)
                else:
                    conceptsUsed.add(dim.typedMember.qname)
        for defaultDim, defaultDimMember in self.modelXbrl.qnameDimensionDefaults.items():
            conceptsUsed.add(defaultDim)
            conceptsUsed.add(defaultDimMember)
        for relationshipSetKey in self.relationshipSets:
            relationshipSet = self.modelXbrl.relationshipSet(*relationshipSetKey)
            for rel in relationshipSet.modelRelationships:
                if isinstance(rel.fromModelObject, ModelConcept):
                    conceptsUsed.add(rel.fromModelObject)
                if isinstance(rel.toModelObject, ModelConcept):
                    conceptsUsed.add(rel.toModelObject)
        for qn in (XbrlConst.qnXbrliIdentifier, XbrlConst.qnXbrliPeriod, XbrlConst.qnXbrliUnit):
            conceptsUsed.add(self.modelXbrl.qnameConcepts[qn])
        return conceptsUsed

    def insertDataDictionary(self):
        # separate graph
        # document-> dataTypeSet -> dataType
        # do all schema dataTypeSet vertices

        self.type_id = {}
        self.aspect_id = {}
        self.aspect_proxy_id = {}
        self.roleType_id = {}
        self.arcroleType_id = {}

        if any((not self.document_isNew[modelDocument.uri])
               for modelDocument in self.modelXbrl.urlDocs.values()):
            conceptsUsed = self.conceptsUsed()

        for modelDocument in self.modelXbrl.urlDocs.values():
            self.showStatus("insert DataDictionary " + modelDocument.basename)

            # don't re-output existing documents
            if modelDocument.type == Type.SCHEMA:
                isNewDocument = self.document_isNew[modelDocument.uri]
                modelConcepts = [modelConcept
                                 for modelConcept in self.modelXbrl.qnameConcepts.values()
                                 if modelConcept.modelDocument is modelDocument and
                                    (isNewDocument or modelConcept in conceptsUsed)]
                if isNewDocument:
                    # adding document as new
                    modelTypes = [modelType
                                  for modelType in self.modelXbrl.qnameTypes.values()
                                  if modelType.modelDocument is modelDocument]
                    conceptAspects = []
                    for modelConcept in modelConcepts:
                        conceptAspect = {'_class': 'aspect',
                                         'name': modelConcept.name}
                        if modelConcept.isAbstract:
                            conceptAspect['isAbstract'] = True
                        if modelConcept.periodType:
                            conceptAspect['periodType'] = modelConcept.periodType
                        if modelConcept.balance:
                            conceptAspect['balance'] = modelConcept.balance
                        for propertyName in ('isItem', 'isTuple', 'isLinkPart', 'isNumeric', 'isMonetary',
                                             'isExplicitDimension', 'isTypedDimension', 'isDomainMember', 'isHypercubeItem',
                                             'isShares', 'isTextBlock'):
                            propertyValue = getattr(modelConcept, propertyName, None)
                            if propertyValue:
                                conceptAspect[propertyName] = propertyValue
                        conceptAspects.append(conceptAspect)
                    roleTypes = [modelRoleType
                                 for modelRoleTypes in self.modelXbrl.roleTypes.values()
                                 for modelRoleType in modelRoleTypes]
                    arcroleTypes = [modelRoleType
                                 for modelRoleTypes in self.modelXbrl.arcroleTypes.values()
                                 for modelRoleType in modelRoleTypes]
                    activity = "Insert data dictionary types, aspects, roles, and arcroles for " + modelDocument.uri
                    results = self.execute(activity, "//" + activity + """
                        reportV = g.v(report_id)
                        docV = g.v(document_id)
                        // add dictV if it doesn't exist
                        dictIt = docV.out('doc_data_dictionary')
                        dictV = (dictIt.hasNext() ? dictIt.next() : g.addVertex(dict) )
                        // add edge from reportV to dictV if not present
                        vReportIn = reportV.in('report_data_dictionary').has('id',reportV.id)
                        vReportIn.hasNext() ?: g.addEdge(reportV, dictV, 'report_data_dictionary')
                        // add edge from docV to dictV if not present
                        vDictIn = dictV.in('doc_data_dictionary').has('id',docV.id)
                        vDictIn.hasNext() ?: g.addEdge(docV, dictV, 'doc_data_dictionary')
                        type_ids = []
                        types.each{
                            typeV = g.addVertex(it)
                            type_ids << typeV.id
                            g.addEdge(dictV, typeV, 'data_type')
                        }
                        aspectsV = g.addVertex(['_class':'aspects'])
                        g.addEdge(dictV, aspectsV, 'aspects')
                        aspect_ids = []
                        aspects.each{
                            aspectV = g.addVertex(it)
                            aspect_ids << aspectV.id
                            g.addEdge(dictV, aspectV, 'aspect')
                            g.addEdge(aspectsV, aspectV, it.name.hashCode().toString() )
                        }
                        role_type_ids = []
                        roletypes.each{
                            roleTypeV = g.addVertex(it)
                            role_type_ids << roleTypeV.id
                            g.addEdge(docV, roleTypeV, 'role_type')
                        }
                        arcrole_type_ids = []
                        arcroletypes.each{
                            arcroleTypeV = g.addVertex(it)
                            arcrole_type_ids << arcroleTypeV.id
                            g.addEdge(docV, arcroleTypeV, 'arcrole_type')
                        }
                        [dictV.id, type_ids, aspect_ids, role_type_ids, arcrole_type_ids]
                        """,
                        params={
                        'report_id': self.report_id,
                        'document_id': self.document_ids[modelDocument.uri],
                        'dict': {
                            '_class': 'data_dictionary',
                            'namespace': modelDocument.targetNamespace},
                        'types': [{
                            '_class': 'data_type',
                            'name': modelType.name
                              } for modelType in modelTypes],
                        'aspects': conceptAspects,
                        'roletypes': [{
                            '_class': 'role_type',
                            'uri': modelRoleType.roleURI,
                            'definition': modelRoleType.definition or ''
                              } for modelRoleType in roleTypes],
                        'arcroletypes': [{
                            '_class': 'arcrole_type',
                            'uri': modelRoleType.arcroleURI,
                            'definition': modelRoleType.definition or '',
                            'cyclesAllowed': modelRoleType.cyclesAllowed
                              } for modelRoleType in arcroleTypes],
                        })["results"]
                    dict_id, type_ids, aspect_ids, role_type_ids, arcrole_type_ids = results
                    self.dict_id = int(dict_id)
                    for iT, type_id in enumerate(type_ids):
                        self.type_id[modelTypes[iT].qname] = int(type_id)
                    for iC, aspect_id in enumerate(aspect_ids):
                        self.aspect_id[modelConcepts[iC].qname] = int(aspect_id)
                    for iRT, roleType_id in enumerate(role_type_ids):
                        self.roleType_id[roleTypes[iRT].roleURI] = int(roleType_id)
                    for iAT, arcroleType_id in enumerate(arcrole_type_ids):
                        self.arcroleType_id[arcroleTypes[iAT].arcroleURI] = int(arcroleType_id)
                    '''
                    results = self.execute("Insert data dictionary types, and arcroles for " +
                                           modelDocument.uri, """
                        reportV = g.v(report_id)
                        docV = g.v(document_id)
                        // add dictV if it doesn't exist
                        dictIt = docV.out('doc_data_dictionary')
                        dictV = (dictIt.hasNext() ? dictIt.next() : g.addVertex(dict) )
                        // add edge from reportV to dictV if not present
                        vReportIn = reportV.in('report_data_dictionary').has('id',reportV.id)
                        vReportIn.hasNext() ?: g.addEdge(reportV, dictV, 'report_data_dictionary')
                        // add edge from docV to dictV if not present
                        vDictIn = dictV.in('doc_data_dictionary').has('id',docV.id)
                        vDictIn.hasNext() ?: g.addEdge(docV, dictV, 'doc_data_dictionary')
                        type_ids = []
                        types.each{
                            typeV = g.addVertex(it)
                            type_ids << typeV.id
                            g.addEdge(dictV, typeV, 'data_type')
                        }
                        arcrole_type_ids = []
                        arcroletypes.each{
                            arcroleTypeV = g.addVertex(it)
                            arcrole_type_ids << arcroleTypeV.id
                            g.addEdge(docV, arcroleTypeV, 'arcrole_type')
                        }
                        [dictV.id, type_ids, arcrole_type_ids]
                        """,
                        params={
                        'report_id': self.report_id,
                        'document_id': self.document_ids[modelDocument.uri],
                        'dict': {
                            '_class': 'data_dictionary',
                            'namespace': modelDocument.targetNamespace},
                        'types': [{
                            '_class': 'data_type',
                            'name': modelType.name
                              } for modelType in modelTypes],
                        'arcroletypes': [{
                            '_class': 'arcrole_type',
                            'uri': modelRoleType.arcroleURI,
                            'definition': modelRoleType.definition or '',
                            'cyclesAllowed': modelRoleType.cyclesAllowed
                              } for modelRoleType in arcroleTypes],
                        })["results"]
                    dict_id, type_ids, arcrole_type_ids = results
                    self.dict_id = int(dict_id)
                    for iT, type_id in enumerate(type_ids):
                        self.type_id[modelTypes[iT].qname] = int(type_id)
                    for iAT, arcroleType_id in enumerate(arcrole_type_ids):
                        self.arcroleType_id[arcroleTypes[iAT].arcroleURI] = int(arcroleType_id)

                    results = self.execute("Insert data dictionary roles for " +
                                           modelDocument.uri, """
                        reportV = g.v(report_id)
                        docV = g.v(document_id)
                        // add dictV if it doesn't exist
                        dictIt = docV.out('doc_data_dictionary')
                        dictV = (dictIt.hasNext() ? dictIt.next() : g.addVertex(dict) )
                        // add edge from reportV to dictV if not present
                        vReportIn = reportV.in('report_data_dictionary').has('id',reportV.id)
                        vReportIn.hasNext() ?: g.addEdge(reportV, dictV, 'report_data_dictionary')
                        // add edge from docV to dictV if not present
                        vDictIn = dictV.in('doc_data_dictionary').has('id',docV.id)
                        vDictIn.hasNext() ?: g.addEdge(docV, dictV, 'doc_data_dictionary')
                        role_type_ids = []
                        roletypes.each{
                            roleTypeV = g.addVertex(it)
                            role_type_ids << roleTypeV.id
                            g.addEdge(docV, roleTypeV, 'role_type')
                        }
                        role_type_ids
                        """,
                        params={
                        'report_id': self.report_id,
                        'document_id': self.document_ids[modelDocument.uri],
                        'dict': {
                            '_class': 'data_dictionary',
                            'namespace': modelDocument.targetNamespace},
                        'roletypes': [{
                            '_class': 'role_type',
                            'uri': modelRoleType.roleURI,
                            'definition': modelRoleType.definition or ''
                              } for modelRoleType in roleTypes],
                        })["results"]
                    role_type_ids = results
                    for iRT, roleType_id in enumerate(role_type_ids):
                        self.roleType_id[roleTypes[iRT].roleURI] = int(roleType_id)


                    results = self.execute("Insert data dictionary aspects for " +
                                           modelDocument.uri, """
                        reportV = g.v(report_id)
                        docV = g.v(document_id)
                        // add dictV if it doesn't exist
                        dictIt = docV.out('doc_data_dictionary')
                        dictV = (dictIt.hasNext() ? dictIt.next() : g.addVertex(dict) )
                        // add edge from reportV to dictV if not present
                        vReportIn = reportV.in('report_data_dictionary').has('id',reportV.id)
                        vReportIn.hasNext() ?: g.addEdge(reportV, dictV, 'report_data_dictionary')
                        // add edge from docV to dictV if not present
                        vDictIn = dictV.in('doc_data_dictionary').has('id',docV.id)
                        vDictIn.hasNext() ?: g.addEdge(docV, dictV, 'doc_data_dictionary')
                        aspectsV = g.addVertex(['_class':'aspects'])
                        g.addEdge(dictV, aspectsV, 'aspects')
                        aspect_ids = []
                        aspects.each{
                            aspectV = g.addVertex(it)
                            aspect_ids << aspectV.id
                            g.addEdge(dictV, aspectV, 'aspect')
                            g.addEdge(aspectsV, aspectV, it.name.hashCode().toString() )
                        }
                        aspect_ids
                        """,
                        params={
                        'report_id': self.report_id,
                        'document_id': self.document_ids[modelDocument.uri],
                        'dict': {
                            '_class': 'data_dictionary',
                            'namespace': modelDocument.targetNamespace},
                        'aspects': conceptAspects,
                        })["results"]
                    aspect_ids = results
                    for iC, aspect_id in enumerate(aspect_ids):
                        self.aspect_id[modelConcepts[iC].qname] = int(aspect_id)
                    '''
                else: # not new, just get aspect (concept) id's
                    results = self.execute("Access existing data dictionary types, aspects, roles, and arcroles for " +
                                           modelDocument.uri, """//Access existing data dictionary aspects
                        aspect_ids = []
                        g.v(document_id).out('doc_data_dictionary').out('aspects').each {
                            aspectsV = it // dereference vertex from pipeline
                            aspect_names.each{
                                aspect_name = it
                                aspectsV.out(aspect_name.hashCode().toString()).each{
                                    if (it.name == aspect_name)
                                        aspect_ids << it.id
                                }
                            }
                        }
                        aspect_ids
                        """,
                        params={'document_id': self.document_ids[modelDocument.uri],
                            'aspect_names': [modelConcept.name for modelConcept in modelConcepts],
                        })["results"]
                    for iC, aspect_id in enumerate(results):
                        self.aspect_id[modelConcepts[iC].qname] = int(aspect_id) if aspect_id is not None else None

        typeDerivationEdges = []
        for modelType in self.modelXbrl.qnameTypes.values():
            if self.document_isNew[modelType.modelDocument.uri]:
                qnamesDerivedFrom = modelType.qnameDerivedFrom
                if not isinstance(qnamesDerivedFrom, (list,tuple)): # list if a union
                    qnamesDerivedFrom = (qnamesDerivedFrom,)
                for qnameDerivedFrom in qnamesDerivedFrom:
                    if modelType.qname in self.type_id and qnameDerivedFrom in self.type_id:
                        typeDerivationEdges.append({
                                'from_id': self.type_id[modelType.qname],
                                'to_id': self.type_id[qnameDerivedFrom],
                                'rel': "derived_from"})
        ### was ### g.addEdge(g.v(it.from_id), g.v(it.to_id), it.rel)
        self.execute("Insert type derivation edges", """
            e.each{
                fromV = g.v(it.from_id)
                toV = g.v(it.to_id)
                vOutIt = fromV.out(it.rel).has('id',toV.id)
                vOutIt.hasNext() ?: g.addEdge(fromV, toV, it.rel)
            }
            """,
            params={'e': typeDerivationEdges})
        aspectEdges = []
        for modelConcept in self.modelXbrl.qnameConcepts.values():
            if self.document_isNew[modelConcept.modelDocument.uri]:
                if modelConcept.qname in self.aspect_id:
                    if modelConcept.typeQname in self.type_id:
                        aspectEdges.append({'from_id': self.aspect_id[modelConcept.qname],
                                            'to_id': self.type_id[modelConcept.typeQname],
                                            'rel': "data_type"})
                    if modelConcept.substitutesForQname in self.type_id:
                        aspectEdges.append({'from_id': self.aspect_id[modelConcept.qname],
                                            'to_id': self.type_id[modelConcept.substitutesForQname.typeQname],
                                            'rel': "substitutes_for"})
                    baseXbrliTypeQnames = modelConcept.baseXbrliTypeQname # may be union or single
                    if not isinstance(baseXbrliTypeQnames, (list,tuple)):
                        baseXbrliTypeQnames = (baseXbrliTypeQnames,) # was single base type
                    for baseXbrliTypeQname in baseXbrliTypeQnames:
                        if baseXbrliTypeQname in self.type_id:
                            aspectEdges.append({'from_id': self.aspect_id[modelConcept.qname],
                                                'to_id': self.type_id[baseXbrliTypeQname],
                                                'rel': "base_xbrli_type"})
        self.execute("Insert aspect edges for data type, substitutes for, and base xbrli type", """
            e.each{
                fromV = g.v(it.from_id)
                toV = g.v(it.to_id)
                vOutIt = fromV.out(it.rel).has('id',toV.id)
                vOutIt.hasNext() ?: g.addEdge(fromV, toV, it.rel)
            }
        """,
        params={'e': aspectEdges})

    '''
    def insertValidCombinations(self):
        # document-> validCombinationsSet-> cubes
        self.showStatus("insert ValidCombinations")

        drsELRs = set(ELR
                      for arcrole, ELR, linkqname, arcqname in self.modelXbrl.baseSets.values()
                      if arcrole == XbrlConst.all)
        hasHcRels = self.modelXbrl.relationshipSet(XbrlConst.all).modelRelationships
        hcConcepts = set(hasHcRel.toModelObject for hasHcRel in hasHcRels)

        # one cube region per head pri item with multiple cube regions
        for hcConcept in hcConcepts:
            # include any other concepts in this pri item with clean inheritance
            for drsELR in drsELRs:
                # each ELR is another cube region

        for allRel in val.modelXbrl.relationshipSet(XbrlConst.all, ELR)
        drsPriItems(val, fromELR, fromPriItem

        ... this becomes an unweildly large model, don't see a use case for compiling it out
'''
    def insertAspectProxies(self, qnames):
        aspectQnames = [qname
                        for qname in qnames
                        if qname not in self.aspect_proxy_id and qname in self.aspect_id]
        #print ("missing qnames: " + ", ".join(str(q) for q in aspectQnames if q not in self.aspect_id))
        results = self.execute("Insert aspect proxies", """
            reportV = g.v(report_id)
            aspectProxyV_ids = []
            aspect_ids.each{
                aspectV = g.v(it)
                aspectProxyV = g.addVertex(['_class':'aspect_proxy'])
                aspectProxyV_ids << aspectProxyV.id
                g.addEdge(aspectV, aspectProxyV, 'proxy')
                g.addEdge(reportV, aspectProxyV, 'report_aspect_proxy')
            }
            aspectProxyV_ids
            """,
            params={'report_id': self.report_id,
                    'aspect_ids': [self.aspect_id[qname] for qname in aspectQnames]}
            )["results"]
        for i, proxy_id in enumerate(results):
            self.aspect_proxy_id[aspectQnames[i]] = proxy_id

    def periodAspectValue(self, context):
        if context.isForeverPeriod:
            return 'forever'
        if context.isInstantPeriod:
            return (str(context.instantDatetime),)
        return (str(context.startDatetime),str(context.endDatetime))

    def insertDataPoints(self):
        # separate graph
        # document-> dataTypeSet -> dataType
        self.showStatus("insert DataPoints")

        # do all schema element vertices
        dataPointObjectIndices = []
        # note these initial aspects Qnames used also must be in conceptsUsed above
        aspectQnamesUsed = {XbrlConst.qnXbrliIdentifier, XbrlConst.qnXbrliPeriod, XbrlConst.qnXbrliUnit}
        dimensions = [] # index by hash of dimension
        if self.modelXbrl.modelDocument.type in (Type.INSTANCE, Type.INLINEXBRL):
            instanceDocument = self.modelXbrl.modelDocument
            dataPoints = []
            entityIdentifiers = [] # index by (scheme, identifier)
            periods = []  # index by (instant,) or (start,end) dates
            units = []  # index by measures (qnames set)
            for fact in self.modelXbrl.factsInInstance:
                aspectQnamesUsed.add(fact.concept.qname)
                dataPointObjectIndices.append(fact.objectIndex)
                datapoint = {'_class': 'data_point',
                             #'name': str(fact.qname), not needed, get from aspect (concept)
                             'source_line': fact.sourceline}
                datapoint['xml_id'] = XmlUtil.elementFragmentIdentifier(fact)
                if fact.context is not None:
                    datapoint['context'] = fact.contextID
                    context = fact.context
                    p = self.periodAspectValue(context)
                    if p not in periods:
                        periods.append(p)
                    e = fact.context.entityIdentifier
                    if e not in entityIdentifiers:
                        entityIdentifiers.append(e)
                    for dimVal in context.qnameDims.values():
                        aspectQnamesUsed.add(dimVal.dimensionQname)
                        if dimVal.isExplicit:
                            aspectQnamesUsed.add(dimVal.memberQname)
                            key = (dimVal.dimensionQname, True, dimVal.memberQname)
                        else:
                            key = (dimVal.dimensionQname, False, dimVal.typedMember.stringValue)
                        if key not in dimensions:
                            dimensions.append(key)
                    if fact.isNumeric:
                        datapoint['effective_value'] = str(fact.effectiveValue)
                        if fact.unit is not None:
                            u = str(fact.unit.measures)  # string for now
                            if u not in units:
                                units.append(u)
                            datapoint['unit']= fact.unitID
                        if fact.precision:
                            datapoint['precision'] = fact.precision
                        if fact.decimals:
                            datapoint['decimals'] = fact.decimals
                    datapoint['value'] = dbString( str(fact.value) ) # compress if very long
                dataPoints.append(datapoint)
            results = self.execute("Insert data points", """
                docV = g.v(document_id)
                dpIt = docV.out('data_points')
                datapointsV = (dpIt.hasNext() ? dpIt.next() : g.addVertex(datapoints_set) )
                dpE = docV.out('data_points').has('id', datapointsV.id)
                dpE.hasNext() ?: g.addEdge(docV, datapointsV, 'data_points')
                datapointV_ids = []
                datapoints.each{
                    dpV = g.addVertex(it)
                    datapointV_ids << dpV.id
                    g.addEdge(datapointsV, dpV, 'data_point')
                }
                [datapointsV.id, datapointV_ids]
                """,
                params={'document_id': self.document_ids[instanceDocument.uri],
                        'datapoints_set': {
                            '_class': 'datapoints_set'},
                        'datapoints': dataPoints}
                )["results"]
            datapointsV_id, datapointVids_list = results
            dataPointVertexIds = dict((dataPointObjectIndices[i], int(id))
                                      for i, id in enumerate(datapointVids_list))

            results = self.execute("Insert entity identifiers", """
                entIdentV_ids = []
                entityIdentifiers.each{
                    entIdentV = g.addVertex(it)
                    entIdentV_ids << entIdentV.id
                }
                entIdentV_ids
                """,
                params={'entityIdentifiers': [{'_class':'entity_identifier',
                                               'scheme': e[0],
                                               'identifier': e[1]}
                                              for e in entityIdentifiers]}
                )["results"]
            entityIdentifierVertexIds = [int(entIdent_id) for entIdent_id in results]

            p = []
            for period in periods:
                if period == 'forever':
                    p.append({'_class': 'period',
                              'forever': 'forever'})
                elif len(period) == 1:
                    p.append({'_class': 'period',
                              'instant': period[0]})
                else:
                    p.append({'_class': 'period',
                              'start_date': period[0],
                              'end_date': period[1]})
            results = self.execute("Insert periods", """
                periodV_ids = []
                periods.each{
                    periodV = g.addVertex(it)
                    periodV_ids << periodV.id
                }
                periodV_ids
                """,
                params={'periods': p}
                )["results"]
            periodVertexIds = [int(period_id) for period_id in results]

            results = self.execute("Insert units", """
                unitV_ids = []
                units.each{
                    unitV = g.addVertex(it)
                    unitV_ids << unitV.id
                }
                unitV_ids
                """,
                params={'units': [{'_class':'unit',
                                   'measures': u}
                                  for u in units]}
                )["results"]
            unitVertexIds = [int(unit_id) for unit_id in results]

            if dimensions:
                self.showStatus("insert aspect value selection groups")
                aspValSels = []
                for dimQn, isExplicit, value in dimensions:
                    if isExplicit:
                        aspValSels.append({'_class': 'aspect_value_selection',
                                           'name':dimQn.localName + '-' + value.localName})
                    else:
                        aspValSels.append({'_class': 'aspect_value_selection',
                                           'name': dimQn.localName + '-' + str(len(aspValSels)+1),
                             '             typed_value': value})

                results = self.execute("Insert aspect value selection groups", """
                    aspectValSelGroupV = g.addVertex(aspect_val_sel_group)
                    aspectValSelV_ids = []
                    aspect_val_sels.each{
                        aspectValSelV = g.addVertex(it)
                        aspectValSelV_ids << aspectValSelV.id
                        g.addEdge(aspectValSelGroupV, aspectValSelV, 'aspect_value_selection_group')
                    }
                    [aspectValSelGroupV.id, aspectValSelV_ids]
                    """,
                    params={'aspect_val_sel_group': {'_class': 'aspect_value_selection_group'},
                            'aspect_val_sels': aspValSels}
                    )["results"]
                aspValSelGrpV_id, aspValSelV_ids_list = results
                aspValSelVertexIds = [int(aspValSel_id) for aspValSel_id in aspValSelV_ids_list]

            else:
                aspValSelVertexIds = []
            dimValAspValSelVertexIds = dict((dimensions[i], aspValSel_id)
                                            for i, aspValSel_id in enumerate(aspValSelVertexIds))

        self.showStatus("insert aspect proxies")
        self.insertAspectProxies(aspectQnamesUsed)

        if dimensions:
            self.showStatus("insert dimension member edges")
            # connect aspectValueSelection to concept dimension and member concepts
            self.execute("Insert dimension member edges", """
                aspects.each{
                    g.addEdge(g.v(it.aspValSel_id), g.v(it.dimension_id), 'aspect')
                }
                aspect_values.each{
                    g.addEdge(g.v(it.aspValSel_id), g.v(it.member_id), 'aspect_value')
                }
                []
                """,
                params={'aspects': [{
                            'aspValSel_id': aspValSel_id,
                            'dimension_id': self.aspect_proxy_id[dimQn]}
                            for i, aspValSel_id in enumerate(aspValSelVertexIds)
                            for dimQn,isExplicit,memQn in dimensions[i:i+1]],
                        'aspect_values': [{
                            'aspValSel_id': aspValSel_id,
                            'member_id': self.aspect_proxy_id[memQn]}
                            for i, aspValSel_id in enumerate(aspValSelVertexIds)
                            for dimQn,isExplicit,memQn in dimensions[i:i+1]
                            if isExplicit]}
                )["results"]

        # add aspect proxy relationships
        edges = []
        if self.modelXbrl.modelDocument.type in (Type.INSTANCE, Type.INLINEXBRL):
            # aspect value - aspect relationships
            for aspectProxyId, rel, aspectValueVertexIds in (
                (self.aspect_proxy_id[XbrlConst.qnXbrliIdentifier], 'entity_identifier_aspects', entityIdentifierVertexIds),
                (self.aspect_proxy_id[XbrlConst.qnXbrliPeriod], 'period_aspects', periodVertexIds),
                (self.aspect_proxy_id[XbrlConst.qnXbrliUnit], 'unit_aspects', unitVertexIds) ):
                for aspectValueVertexId in aspectValueVertexIds:
                    edges.append({'from_id': aspectValueVertexId,
                                  'to_id': aspectProxyId,
                                  'rel': rel})
        # fact - aspect relationships
        for i, factObjectIndex in enumerate(dataPointObjectIndices):
            fact =  self.modelXbrl.modelObjects[factObjectIndex]
            dataPoint_id = dataPointVertexIds[factObjectIndex]
            # fact concept aspect
            edges.append({
                'from_id': dataPoint_id,
                'to_id': self.aspect_proxy_id[fact.qname],
                'rel': "base_item"})
            context = fact.context
            if context is not None:
                # entityIdentifier aspect
                edges.append({
                    'from_id': dataPoint_id,
                    'to_id': entityIdentifierVertexIds[entityIdentifiers.index(context.entityIdentifier)],
                    'rel': "entity_identifier"})
                # period aspect
                edges.append({
                    'from_id': dataPoint_id,
                    'to_id': periodVertexIds[periods.index(self.periodAspectValue(context))],
                    'rel': "period"})
                # dimension aspectValueSelections
                for dimVal in context.qnameDims.values():
                    key = (dimVal.dimensionQname, dimVal.isExplicit,
                           dimVal.memberQname if dimVal.isExplicit else dimVal.typedMember.stringValue)
                    edges.append({
                        'from_id': dataPoint_id,
                        'to_id': dimValAspValSelVertexIds[key],
                        'rel': "aspect_value_selection"})
            if fact.isNumeric and fact.unit is not None:
                # unit aspect
                u = str(fact.unit.measures)  # string for now
                edges.append({
                    'from_id': dataPoint_id,
                    'to_id': unitVertexIds[units.index(u)],
                    'rel': "_unit"})
            for tupleFact in fact.modelTupleFacts:
                # edge to tuple from item
                edges.append({
                    'from_id': dataPointVertexIds[tupleFact.objectIndex],
                    'to_id': dataPoint_id,
                    'rel': "tuple"})

        self.showStatus("insert aspect relationship edges")
        results = self.execute("Insert aspect relationship edges", """
            e.each{g.addEdge(g.v(it.from_id), g.v(it.to_id), it.rel)}
            []
            """,
            params={'e': edges})["results"]

    def insertRelationshipSets(self):
        self.showStatus("insert relationship sets")
        results = self.execute("Insert relationship sets", """
            reportV = g.v(report_id)
            relSetsV = g.addVertex(relSets)
            g.addEdge(reportV, relSetsV, 'relationship_sets')
            relSetV_ids = []
            relSet.each{
                relSetV = g.addVertex(it)
                relSetV_ids << relSetV.id
                g.addEdge(relSetsV, relSetV, 'relationship_set')}
            [relSetsV.id, relSetV_ids]
            """,
            params={
                'report_id': self.report_id,
                'relSets': {
                    '_class': 'relationship_sets'},
                'relSet': [{
                    '_class': 'relationship_set',
                    'arcrole': arcrole,
                    'linkrole': linkrole,
                    'linkdefinition': self.modelXbrl.roleTypeDefinition(linkrole) or '',
                    'linkname': str(linkqname),
                    'arcname': str(arcqname)
                    } for arcrole, linkrole, linkqname, arcqname in self.relationshipSets]
            })["results"]
        relSetsV_id, relSetV_ids_list = results
        relationshipSetIDs = [int(relSet_id) for relSet_id in relSetV_ids_list]

        # do tree walk to build relationships with depth annotated, no targetRole navigation
        relE = [] # fromV, toV, label
        resources = set()
        aspectQnamesUsed = set()
        resourceIDs = {} # index by object

        def walkTree(rels, seq, depth, relationshipSet, visited, relationshipSetId, doVertices):
            for rel in rels:
                if rel not in visited:
                    visited.add(rel)

                    if not doVertices:
                        _relProp = {'seq':seq,
                                    'depth':depth,
                                    '_order':rel.order,
                                    'priority':rel.priority,
                                    'rel_set':relationshipSetId}
                    if isinstance(rel.fromModelObject, ModelConcept):
                        if doVertices:
                            aspectQnamesUsed.add(rel.fromModelObject.qname)
                            sourceId = True
                        else:
                            sourceId = self.aspect_proxy_id[rel.fromModelObject.qname]
                    else:
                        sourceId = None # tbd
                    toModelObject = rel.toModelObject
                    if isinstance(toModelObject, ModelConcept):
                        if doVertices:
                            aspectQnamesUsed.add(toModelObject.qname)
                            targetId = True
                        else:
                            targetId = self.aspect_proxy_id[toModelObject.qname]
                    elif isinstance(toModelObject, ModelResource):
                        if doVertices:
                            resources.add(toModelObject)
                            targetId = 0 # just can't be None, but doesn't matter on doVertices pass
                        else:
                            if rel.preferredLabel:
                                _relProp['preferred_label'] = rel.preferredLabel
                            if rel.arcrole in (XbrlConst.all, XbrlConst.notAll):
                                _relProp['cube_closed'] = rel.closed
                            elif rel.arcrole in (XbrlConst.dimensionDomain, XbrlConst.domainMember):
                                _relProp['aspect_value_usable'] = rel.usable
                            elif rel.arcrole == XbrlConst.summationItem:
                                _relProp['weight'] = rel.weight
                            if relationshipSet.arcrole == "XBRL-dimensions":
                                _relProp['arcrole'] = os.path.basename(rel.arcrole)
                            if toModelObject.role:
                                _relProp['resource_role'] = toModelObject.role
                            targetId = resourceIDs[toModelObject]
                    else:
                        targetId = None # tbd
                    if sourceId is not None and targetId is not None:
                        targetRelationshipSetId = relationshipSetId
                        if relationshipSet.arcrole == "XBRL-dimensions" and rel.targetRole:
                            targetRelSet = self.modelXbrl.relationshipSet(relationshipSet.arcrole, rel.targetRole)
                            for i, relationshipSetKey in enumerate(self.relationshipSets):
                                arcrole, ELR, linkqname, arcqname = relationshipSetKey
                                if arcrole == "XBRL-dimensions" and ELR == rel.targetRole:
                                    targetRelationshipSetId = relationshipSetIDs[i]
                                    break
                            if not doVertices:
                                _relProp['target_linkrole'] = rel.targetRole
                                _relProp['target_rel_set'] = targetRelationshipSetId
                        else:
                            targetRelSet = relationshipSet
                        if doVertices:
                            thisRelId = 0
                        else:
                            relE.append({'from_id': sourceId, 'to_id': targetId, 'label': 'rel', 'properties': _relProp})
                        seq += 1
                        seq = walkTree(targetRelSet.fromModelObject(toModelObject), seq, depth+1, relationshipSet, visited, targetRelationshipSetId, doVertices)
                    visited.remove(rel)
            return seq

        for doVertices in range(1,-1,-1):  # pass 0 = vertices, pass 1 = edges
            for i, relationshipSetKey in enumerate(self.relationshipSets):
                arcrole, ELR, linkqname, arcqname = relationshipSetKey
                relationshipSetId = relationshipSetIDs[i]
                relationshipSet = self.modelXbrl.relationshipSet(arcrole, ELR, linkqname, arcqname)
                seq = 1
                for rootConcept in relationshipSet.rootConcepts:
                    if not doVertices:
                        aspectId = self.aspect_proxy_id[rootConcept.qname]
                        relE.append({'from_id': relationshipSetId, 'to_id': aspectId, 'label': 'root'})
                    seq = walkTree(relationshipSet.fromModelObject(rootConcept), seq, 1, relationshipSet, set(), relationshipSetId, doVertices)
            if doVertices:
                if resources:
                    resourceV = []
                    resourceObjs = []
                    for resource in resources:
                        resourceParam = {'_class': resource.localName,
                                         'value': dbString( resource.stringValue )} # compress if very long
                        if resource.role:
                            resourceParam['role'] = resource.role
                        resourceV.append(resourceParam)
                        resourceObjs.append(resource) # need these in a list in same order as resoureV
                    for i, v_id in enumerate(self.execute("Insert relationship set concept-to-resource relationships", """
                        resourceV_ids = []
                        resourceV.each{resourceV_ids << g.addVertex(it).id}
                        resourceV_ids
                        """,
                        params={'resourceV': resourceV}
                        )["results"]):
                        resourceIDs[resourceObjs[i]] = int(v_id)

                self.insertAspectProxies(aspectQnamesUsed)
            else:
                self.execute("Insert relationship edges", """
                    relE.each{
                        if (it.properties) {
                            g.addEdge(g.v(it.from_id), g.v(it.to_id), it.label, it.properties)
                        } else {
                            g.addEdge(g.v(it.from_id), g.v(it.to_id), it.label)
                        }
                    }
                    []
                    """,
                    params={'relE': relE}
                    )["results"]

                # TBD: do we want to link resources to the report (by role, class, or otherwise?)

        resourceIDs.clear() # dereferemce objects
        resources = None

    def insertValidationResults(self):
        logEntries = []
        for handler in logging.getLogger("arelle").handlers:
            if hasattr(handler, "dbHandlerLogEntries"):
                logEntries = handler.dbHandlerLogEntries()
                break

        messages = []
        messageRefs = [] # direct link to objects
        for i, logEntry in enumerate(logEntries):
            # capture message ref's
            msgRefIds = []
            for ref in logEntry['refs']:
                modelObject = self.modelXbrl.modelObject(ref.get('objectId',''))
                # for now just find a concept
                aspectQname = None
                if isinstance(modelObject, (ModelConcept, ModelFact)):
                    aspectQname = modelObject.qname
                elif isinstance(modelObject, ModelRelationship):
                    if isinstance(modelObject.toModelObject, ModelConcept):
                        aspectQname = modelObject.toModelObject.qname
                    elif isinstance(modelObject.fromModelObject, ModelConcept):
                        aspectQname = modelObject.fromModelObject.qname
                if aspectQname is not None and aspectQname in self.aspect_proxy_id:
                    msgRefIds.append(self.aspect_proxy_id[aspectQname])


            messages.append({'_class': 'message',
                             'seq': i + 1,
                             'code': logEntry['code'],
                             'level': logEntry['level'],
                             'text': dbString( logEntry['message']['text'] ),
                             'refs': msgRefIds})
        if messages:
            self.showStatus("insert validation messages")
            results = self.execute("Insert validation messages", """
                filingV = g.v(filing_id)
                msgsV = g.addVertex(['_class':'messages'])
                g.addEdge(filingV, msgsV, 'validation_messages')
                msgV_ids = []
                messages.each{
                    msgV = g.addVertex(it.subMap(['_class','seq','code','level','text']))
                    msgV_ids << msgV.id
                    g.addEdge(msgsV, msgV, 'message')
                    it['refs'].each{
                        g.addEdge(msgV, g.v(it), 'message_ref')
                    }}
                msgV_ids
                """,
                params={
                    'filing_id': self.filing_id,
                    'messages': messages
                })["results"]
            relationshipSetIDs = [int(msg_id) for msg_id in results]



'''
Use this module to start Arelle in windowing interactive UI

See COPYRIGHT.md for copyright information.
'''
import sys

if sys.platform == "darwin" and getattr(sys, 'frozen', False):
    for i in range(len(sys.path)): # signed code can't contain python modules
        sys.path.append(sys.path[i].replace("MacOS", "Resources"))
        
from arelle import CntlrWinMain

CntlrWinMain.main()



'''
StreamingExtensions is a plug-in to both GUI menu and command line/web service
that provides an alternative approach to big instance documents without building a DOM, to save
memory footprint.  lxml iterparse is used to parse the big instance.  ModelObjects are specialized by features
for efficiency and to avoid dependency on an underlying DOM.

(Note that this module is based on iterparse, the module under the installation/plugs is much faster.)

See COPYRIGHT.md for copyright information.
'''

import io, sys, os, time
from decimal import Decimal, InvalidOperation
from lxml import etree
from collections import defaultdict
from arelle import XbrlConst, XmlUtil, XmlValidate, ValidateXbrlDimensions
from arelle.ModelDocument import ModelDocument, Type
from arelle.ModelObject import ModelObject
from arelle.ModelObjectFactory import parser
from arelle.ModelValue import QName
from arelle.ModelInstanceObject import ModelContext, ModelFact, ModelUnit
from arelle.Validate import Validate
from arelle.Version import authorLabel, copyrightLabel

_streamingExtensionsValidate = False
_streamingExtensionsCheck = False

def precedingProcessingInstruction(elt, target):
    pi = elt.getprevious()
    while pi is not None:
        if isinstance(pi, etree._ProcessingInstruction) and pi.target == target:
            return pi
        pi = pi.getprevious()
    return None

def streamingExtensionsLoader(modelXbrl, mappedUri, filepath):
    # check if big instance and has header with an initial incomplete tree walk (just 2 elements
    def logSyntaxErrors(parsercontext):
        for error in parsercontext.error_log:
            modelXbrl.error("xmlSchema:syntax",
                    _("%(error)s, %(fileName)s, line %(line)s, column %(column)s, %(sourceAction)s source element"),
                    modelObject=modelDocument, fileName=os.path.basename(filepath),
                    error=error.message, line=error.line, column=error.column, sourceAction="streaming")
    #### note: written for iterparse of lxml prior to version 3.3, otherwise rewrite to use XmlPullParser ###
    #### note: iterparse wants a binary file, but file is text mode
    _file, = modelXbrl.fileSource.file(filepath, binary=True)
    startedAt = time.time()
    modelXbrl.profileActivity()
    parsercontext = etree.iterparse(_file, events=("start","end"), huge_tree=True)
    foundInstance = False
    foundErrors = False
    streamingAspects = None
    numRootFacts1 = 0
    numElts = 0
    elt = None
    for event, elt in parsercontext:
        if event == "start":
            if elt.getparent() is not None:
                if elt.getparent().tag == "{http://www.xbrl.org/2003/instance}xbrl":
                    if not foundInstance:
                        foundInstance = True
                        pi = precedingProcessingInstruction(elt, "xbrl-streamable-instance")
                        if pi is None:
                            break
                        else:
                            streamingAspects = dict(pi.attrib.copy())
                    if not elt.tag.startswith("{http://www.xbrl.org/"):
                        numRootFacts1 += 1
                        if numRootFacts1 % 1000 == 0:
                            modelXbrl.profileActivity("... streaming tree check", minTimeToShow=20.0)
                elif not foundInstance:
                    break
            elif elt.tag == "{http://www.xbrl.org/2003/instance}xbrl" and precedingProcessingInstruction(elt, "xbrl-streamable-instance") is not None:
                modelXbrl.error("streamingExtensions:headerMisplaced",
                        _("Header is misplaced: %(error)s, must follow xbrli:xbrl element"),
                        modelObject=elt)
        elif event == "end":
            elt.clear()
            numElts += 1
            if numElts % 1000 == 0 and elt.getparent() is not None:
                while elt.getprevious() is not None and elt.getparent() is not None:
                    del elt.getparent()[0]
    if elt is not None:
        elt.clear()
    _file.seek(0,io.SEEK_SET) # allow reparsing
    if not foundInstance or streamingAspects is None:
        del elt, parsercontext
        _file.close()
        return None
    modelXbrl.profileStat(_("streaming tree check"), time.time() - startedAt)
    startedAt = time.time()
    try:
        version = Decimal(streamingAspects.get("version"))
        if int(version) != 1:
            modelXbrl.error("streamingExtensions:unsupportedVersion",
                    _("Streaming version %(version)s, major version number must be 1"),
                    modelObject=elt, version=version)
            foundErrors = True
    except (InvalidOperation, OverflowError):
        modelXbrl.error("streamingExtensions:versionError",
                _("Version %(version)s, number must be 1.n"),
                modelObject=elt, version=streamingAspects.get("version", "(none)"))
        foundErrors = True
    for bufAspect in ("contextBuffer", "unitBuffer", "footnoteBuffer"):
        try:
            bufLimit = Decimal(streamingAspects.get(bufAspect, "INF"))
            if bufLimit < 1 or (bufLimit.is_finite() and bufLimit % 1 != 0):
                raise InvalidOperation
            elif bufAspect == "contextBuffer":
                contextBufferLimit = bufLimit
            elif bufAspect == "unitBuffer":
                unitBufferLimit = bufLimit
            elif bufAspect == "footnoteBuffer":
                footnoteBufferLimit = bufLimit
        except InvalidOperation:
            modelXbrl.error("streamingExtensions:valueError",
                    _("Streaming %(attrib)s %(value)s, number must be a positive integer or INF"),
                    modelObject=elt, attrib=bufAspect, value=streamingAspects.get(bufAspect))
            foundErrors = True
    if parsercontext.error_log:
        foundErrors = True
    logSyntaxErrors(parsercontext)

    if foundErrors:
        _file.close()
        return None
    parsercontext = etree.iterparse(_file, events=("start","end"), huge_tree=True)
    _parser, _parserLookupName, _parserLookupClass = parser(modelXbrl,filepath)
    eltMdlObjs = {}
    beforeInstanceStream = True
    validator = None
    contextBuffer = []
    unitBuffer = []
    footnoteBuffer = []
    factBuffer = []
    numFacts = numRootFacts2 = 1
    for event, elt in parsercontext:
        if event == "start":
            mdlObj = _parser.makeelement(elt.tag, attrib=elt.attrib, nsmap=elt.nsmap)
            mdlObj.sourceline = elt.sourceline
            eltMdlObjs[elt] = mdlObj
            if elt.getparent() is None:
                modelDocument = ModelDocument(modelXbrl, Type.INSTANCE, mappedUri, filepath, etree.ElementTree(mdlObj))
                modelDocument.xmlRootElement = mdlObj
                modelXbrl.modelDocument = modelDocument # needed for incremental validation
                mdlObj.init(modelDocument)
                modelXbrl.info("streamingExtensions:streaming",
                               _("Stream processing this instance."),
                               modelObject = modelDocument)
            else:
                eltMdlObjs[elt.getparent()].append(mdlObj)
                mdlObj._init()
                ns = mdlObj.namespaceURI
                ln = mdlObj.localName
                if (beforeInstanceStream and (
                    (ns == XbrlConst.link and ln not in ("schemaRef", "linkbaseRef")) or
                    (ns == XbrlConst.xbrli and ln in ("context", "unit")) or
                    (ns not in (XbrlConst.link, XbrlConst.xbrli)))):
                    beforeInstanceStream = False
                    if _streamingExtensionsValidate:
                        validator = Validate(modelXbrl)
                        validator.instValidator.validate(modelXbrl, modelXbrl.modelManager.formulaOptions.typedParameters(modelXbrl.prefixedNamespaces))
                    else: # need default dimensions
                        ValidateXbrlDimensions.loadDimensionDefaults(modelXbrl)
            mdlObj = None # deref

        elif event == "end":
            mdlObj = eltMdlObjs.pop(elt)
            if elt.text: # text available after child nodes processed
                mdlObj.text = elt.text
            ns = mdlObj.namespaceURI
            ln = mdlObj.localName
            parentMdlObj = mdlObj.getparent()
            if ns == XbrlConst.xbrli:
                if ln == "context":
                    if mdlObj.get("sticky"):
                        del mdlObj.attrib["sticky"]
                        modelDocument.contextDiscover(mdlObj)
                    else:
                        if _streamingExtensionsValidate and len(contextBuffer) >= contextBufferLimit:
                            # drop before adding as dropped may have same id as added
                            cntx = contextBuffer.pop(0)
                            dropContext(modelXbrl, cntx)
                            del parentMdlObj[parentMdlObj.index(cntx)]
                            cntx = None
                        modelDocument.contextDiscover(mdlObj)
                        if contextBufferLimit.is_finite():
                            contextBuffer.append(mdlObj)
                    if _streamingExtensionsValidate:
                        contextsToCheck = (mdlObj,)
                        validator.instValidator.checkContexts(contextsToCheck)
                        if modelXbrl.hasXDT:
                            validator.instValidator.checkContextsDimensions(contextsToCheck)
                        del contextsToCheck # dereference
                elif ln == "unit":
                    if _streamingExtensionsValidate and len(unitBuffer) >= unitBufferLimit:
                        # drop before additing as dropped may have same id as added
                        unit = unitBuffer.pop(0)
                        dropUnit(modelXbrl, unit)
                        del parentMdlObj[parentMdlObj.index(unit)]
                        unit = None
                    modelDocument.unitDiscover(mdlObj)
                    if unitBufferLimit.is_finite():
                        unitBuffer.append(mdlObj)
                    if _streamingExtensionsValidate:
                        validator.instValidator.checkUnits( (mdlObj,) )
                elif ln == "xbrl": # end of document
                    # check remaining footnote refs
                    for footnoteLink in footnoteBuffer:
                        checkFootnoteHrefs(modelXbrl, footnoteLink)
                elt.clear()
            elif ns == XbrlConst.link:
                if ln in ("schemaRef", "linkbaseRef"):
                    modelDocument.discoverHref(mdlObj)
                elif ln in ("roleRef", "arcroleRef"):
                    modelDocument.linkbaseDiscover((mdlObj,), inInstance=True)
                elif ln == "footnoteLink":
                    footnoteLinks = (mdlObj,)
                    modelDocument.linkbaseDiscover(footnoteLinks, inInstance=True)
                    if footnoteBufferLimit.is_finite():
                        footnoteBuffer.append(mdlObj)
                    if _streamingExtensionsValidate:
                        validator.instValidator.checkLinks(footnoteLinks)
                        if len(footnoteBuffer) > footnoteBufferLimit:
                            # check that hrefObjects for locators were all satisfied
                                # drop before addition as dropped may have same id as added
                            footnoteLink = footnoteBuffer.pop(0)
                            checkFootnoteHrefs(modelXbrl, footnoteLink)
                            dropFootnoteLink(modelXbrl, footnoteLink)
                            del parentMdlObj[parentMdlObj.index(footnoteLink)]
                            footnoteLink = None
                    footnoteLinks = None
                elt.clear()
            elif parentMdlObj.qname == XbrlConst.qnXbrliXbrl:
                numRootFacts2 += 1
                modelDocument.factDiscover(mdlObj, modelXbrl.facts)
                XmlValidate.validate(modelXbrl, mdlObj)
                if _streamingExtensionsValidate:
                    factsToCheck = (mdlObj,)
                    validator.instValidator.checkFacts(factsToCheck)
                    if modelXbrl.hasXDT:
                        validator.instValidator.checkFactsDimensions(factsToCheck)
                    del factsToCheck
                    dropFact(modelXbrl, mdlObj, modelXbrl.facts)
                    del parentMdlObj[parentMdlObj.index(mdlObj)]
                if numRootFacts2 % 1000 == 0:
                    modelXbrl.profileActivity("... streaming fact {0} of {1} {2:.2f}%".format(numRootFacts2, numRootFacts1, 100.0 * numRootFacts2 / numRootFacts1),
                                              minTimeToShow=20.0)
                # get rid of root element from iterparse's tree
                elt.clear()
                while elt.getprevious() is not None:  # cleans up any prior siblings
                    del elt.getparent()[0]
            mdlObj = None # deref
    logSyntaxErrors(parsercontext)
    del parsercontext
    if validator is not None:
        validator.close()
    _file.close()
    modelXbrl.profileStat(_("streaming complete"), time.time() - startedAt)
    return modelDocument

def checkFootnoteHrefs(modelXbrl, footnoteLink):
    for locElt in footnoteLink.iterchildren(tag="{http://www.xbrl.org/2003/linkbase}loc"):
        for hrefElt, doc, id in footnoteLink.modelDocument.hrefObjects:
            if locElt == hrefElt and id not in footnoteLink.modelDocument.idObjects:
                modelXbrl.error("streamingExtensions:footnoteId",
                        _("Footnote id %(id)s not matched to fact in buffered region"),
                        modelObject=footnoteLink, id=id)

def dropContext(modelXbrl, cntx):
    del modelXbrl.contexts[cntx.id]
    dropObject(modelXbrl, cntx)

def dropUnit(modelXbrl, unit):
    del modelXbrl.units[unit.id]
    dropObject(modelXbrl, unit)

def dropFootnoteLink(modelXbrl, footnoteLink):
    for baseSet in modelXbrl.baseSets.values():
        if footnoteLink in baseSet:
            baseSet.remove(footnoteLink)
    dropObject(modelXbrl, footnoteLink)

def dropFact(modelXbrl, fact, facts):
    while fact.modelTupleFacts:
        dropFact(modelXbrl, fact.modelTupleFacts[0], fact.modelTupleFacts)
    modelXbrl.factsInInstance.discard(fact)
    facts.remove(fact)
    modelXbrl.modelObjects[fact.objectIndex] = None # objects found by index, can't remove position from list
    fact.modelDocument.modelObjects.remove(fact)
    fact.clear()

def dropObject(modelXbrl, mdlObj):
    for childObj in mdlObj.iterchildren():
        dropObject(modelXbrl, childObj)
    if mdlObj.qname == XbrlConst.qnLinkLoc:
        hrefs = mdlObj.modelDocument.hrefObjects
        removedHrefs = [i for i, hrefObj in enumerate(hrefs) if mdlObj == hrefObj[0]]
        for i in removedHrefs:
            del hrefs[i]
    modelXbrl.modelObjects[mdlObj.objectIndex] = None # objects found by index, can't remove position from list
    mdlObj.modelDocument.modelObjects.remove(mdlObj)
    mdlObj.modelDocument.idObjects.pop(mdlObj.id, None)
    mdlObj.clear()

def streamingOptionsExtender(parser):
    parser.add_option("--check-streaming",
                      action="store_true",
                      dest="check_streaming",
                      help=_('Check streamability of instance document."'))

def streamingExtensionsSetup(self, options, **kwargs):
    global _streamingExtensionsCheck, _streamingExtensionsValidate
    _streamingExtensionsCheck = getattr(options, 'check_streaming', False)
    _streamingExtensionsValidate = options.validate
    if options.validate:
        options.validate = False # prevent cmdLine calling validation

'''
   Do not use _( ) in pluginInfo itself (it is applied later, after loading
'''

__pluginInfo__ = {
    'name': 'Streaming Extensions Loader',
    'version': '0.9',
    'description': "This plug-in loads big XBRL instances without building a DOM in memory.  "
                    "lxml iterparse parses XBRL directly into an object model without a DOM.  ",
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'CntlrCmdLine.Options': streamingOptionsExtender,
    'CntlrCmdLine.Utility.Run': streamingExtensionsSetup,
    'ModelDocument.PullLoader': streamingExtensionsLoader,
}



'''
See COPYRIGHT.md for copyright information.
'''
import regex as re
from collections import defaultdict
from arelle import (ModelDocument, ModelRelationshipSet, XmlUtil, XbrlConst)
from arelle.ModelDtsObject import ModelConcept, ModelResource
from arelle.ModelObject import ModelObject
from arelle.UrlUtil import isHttpUrl
from .Dimensions import checkFilingDimensions
from .DTS import checkFilingDTS


def validateFiling(val, modelXbrl):

    linkroleDefinitionStatementSheet = re.compile(r"[^-]+-\s+Statement\s+-\s+.*", # no restriction to type of statement
                                                  re.IGNORECASE)
    if not hasattr(modelXbrl.modelDocument, "xmlDocument"): # not parsed
        return

    val._isStandardUri = {}
    modelXbrl.modelManager.disclosureSystem.loadStandardTaxonomiesDict()

    # find typedDomainRefs before validateXBRL pass
    val.typedDomainQnames = set()
    val.typedDomainElements = set()
    for modelConcept in modelXbrl.qnameConcepts.values():
        if modelConcept.isTypedDimension:
            typedDomainElement = modelConcept.typedDomainElement
            if isinstance(typedDomainElement, ModelConcept):
                val.typedDomainQnames.add(typedDomainElement.qname)
                val.typedDomainElements.add(typedDomainElement)

    # note that some XFM tests are done by ValidateXbrl to prevent multiple node walks
    xbrlInstDoc = modelXbrl.modelDocument.xmlDocument.getroot()
    disclosureSystem = val.disclosureSystem
    disclosureSystemVersion = disclosureSystem.version

    modelXbrl.modelManager.showStatus(_("validating {0}").format(disclosureSystem.name))

    val.modelXbrl.profileActivity()
    conceptsUsed = {} # key=concept object value=True if has presentation label
    labelsRelationshipSet = modelXbrl.relationshipSet(XbrlConst.conceptLabel)
    genLabelsRelationshipSet = modelXbrl.relationshipSet(XbrlConst.elementLabel)
    presentationRelationshipSet = modelXbrl.relationshipSet(XbrlConst.parentChild)
    referencesRelationshipSetWithProhibits = modelXbrl.relationshipSet(XbrlConst.conceptReference, includeProhibits=True)
    val.modelXbrl.profileActivity("... cache lbl, pre, ref relationships", minTimeToShow=1.0)

    val.validateLoggingSemantic = validateLoggingSemantic = (
          modelXbrl.isLoggingEffectiveFor(level="WARNING-SEMANTIC") or
          modelXbrl.isLoggingEffectiveFor(level="ERROR-SEMANTIC"))

    # instance checks
    val.fileNameBasePart = None # prevent testing on fileNameParts if not instance or invalid
    val.fileNameDate = None
    val.entityRegistrantName = None
    val.requiredContext = None
    val.standardNamespaceConflicts = defaultdict(set)
    val.exhibitType = None # e.g., EX-101, EX-201


    # entry point schema checks
    if modelXbrl.modelDocument.type == ModelDocument.Type.SCHEMA:
        # entry must have a P-link
        if not any(hrefElt.localName == "linkbaseRef" and hrefElt.get("{http://www.w3.org/1999/xlink}role") == "http://www.xbrl.org/2003/role/presentationLinkbaseRef"
                   for hrefElt, hrefDoc, hrefId in modelXbrl.modelDocument.hrefObjects):
            modelXbrl.error("SBR.NL.2.2.10.01",
                'Entrypoint schema must have a presentation linkbase', modelObject=modelXbrl.modelDocument)
    # all-labels and references checks
    for concept in modelXbrl.qnameConcepts.values():
        conceptHasDefaultLangStandardLabel = False
        for modelLabelRel in labelsRelationshipSet.fromModelObject(concept):
            modelLabel = modelLabelRel.toModelObject
            role = modelLabel.role
            text = modelLabel.text
            lang = modelLabel.xmlLang
            if role == XbrlConst.documentationLabel:
                if concept.modelDocument.targetNamespace in disclosureSystem.standardTaxonomiesDict:
                    modelXbrl.error("SBR.NL.2.1.0.08",
                        _("Concept %(concept)s of a standard taxonomy cannot have a documentation label: %(text)s"),
                        modelObject=modelLabel, concept=concept.qname, text=text)
            elif text and lang and disclosureSystem.defaultXmlLang and lang.startswith(disclosureSystem.defaultXmlLang):
                if role == XbrlConst.standardLabel:  # merge of pre-plugin code per LOGIUS
                    conceptHasDefaultLangStandardLabel = True
                match = modelXbrl.modelManager.disclosureSystem.labelCheckPattern.search(text)
                if match:
                    modelXbrl.error("SBR.NL.2.3.8.07",
                        'Label for concept %(concept)s role %(role)s has disallowed characters: "%(text)s"',
                        modelObject=modelLabel, concept=concept.qname, role=role, text=match.group())
        for modelRefRel in referencesRelationshipSetWithProhibits.fromModelObject(concept):
            modelReference = modelRefRel.toModelObject
            text = XmlUtil.innerText(modelReference)
            #6.18.1 no reference to company extension concepts
            if (concept.modelDocument.targetNamespace in disclosureSystem.standardTaxonomiesDict and
                not isStandardUri(val, modelRefRel.modelDocument.uri)): # merge of pre-plugin code per LOGIUS
                #6.18.2 no extension to add or remove references to standard concepts
                modelXbrl.error("SBR.NL.2.1.0.08",
                    _("References for standard taxonomy concept %(concept)s are not allowed in an extension linkbase: %(text)s"),
                    modelObject=modelReference, concept=concept.qname, text=text, xml=XmlUtil.xmlstring(modelReference, stripXmlns=True, contentsOnly=True))
        if concept.isItem or concept.isTuple:
            if concept.modelDocument.targetNamespace not in disclosureSystem.standardTaxonomiesDict:
                if not conceptHasDefaultLangStandardLabel:
                    modelXbrl.error("SBR.NL.2.2.2.26",
                        _("Concept %(concept)s missing standard label in local language."),
                        modelObject=concept, concept=concept.qname)
                subsGroup = concept.get("substitutionGroup")
                if ((not concept.isAbstract or subsGroup == "sbr:presentationItem") and
                    not (presentationRelationshipSet.toModelObject(concept) or
                         presentationRelationshipSet.fromModelObject(concept))):
                    modelXbrl.error("SBR.NL.2.2.2.04",
                        _("Concept %(concept)s not referred to by presentation relationship."),
                        modelObject=concept, concept=concept.qname)
                elif ((concept.isDimensionItem or
                      (subsGroup and (subsGroup.endswith(":domainItem") or subsGroup.endswith(":domainMemberItem")))) and
                    not (presentationRelationshipSet.toModelObject(concept) or
                         presentationRelationshipSet.fromModelObject(concept))):
                    modelXbrl.error("SBR.NL.2.2.10.03",
                        _("DTS concept %(concept)s not referred to by presentation relationship."),
                        modelObject=concept, concept=concept.qname)
                if (concept.substitutionGroupQname and
                    concept.substitutionGroupQname.namespaceURI not in disclosureSystem.baseTaxonomyNamespaces):
                    modelXbrl.error("SBR.NL.2.2.2.05",
                        _("Concept %(concept)s has a substitutionGroup of a non-standard concept."),
                        modelObject=concept, concept=concept.qname)

                if concept.isTuple: # verify same presentation linkbase nesting
                    for missingQname in set(concept.type.elements) ^ pLinkedNonAbstractDescendantQnames(modelXbrl, concept):
                        modelXbrl.error("SBR.NL.2.3.4.01",
                            _("Tuple %(concept)s has mismatch between content and presentation children: %(missingQname)s."),
                            modelObject=concept, concept=concept.qname, missingQname=missingQname)
            checkConceptLabels(val, modelXbrl, labelsRelationshipSet, disclosureSystem, concept)
            checkConceptLabels(val, modelXbrl, genLabelsRelationshipSet, disclosureSystem, concept)

    val.modelXbrl.profileActivity("... filer concepts checks", minTimeToShow=1.0)

    # checks on all documents: instance, schema, instance
    checkFilingDTS(val, modelXbrl.modelDocument, [])
    ''' removed RH 2011-12-23, corresponding use of nameWordsTable in ValidateFilingDTS
    if val.validateSBRNL:
        del val.nameWordsTable
    '''
    val.modelXbrl.profileActivity("... filer DTS checks", minTimeToShow=1.0)

    conceptRelsUsedWithPreferredLabels = defaultdict(list)
    usedCalcsPresented = defaultdict(set) # pairs of concepts objectIds used in calc
    usedCalcFromTosELR = {}
    localPreferredLabels = defaultdict(set)
    drsELRs = set()

    # do calculation, then presentation, then other arcroles
    val.summationItemRelsSetAllELRs = modelXbrl.relationshipSet(XbrlConst.summationItem)
    for arcroleFilter in (XbrlConst.summationItem, XbrlConst.parentChild, "*"):
        for baseSetKey, baseSetModelLinks  in modelXbrl.baseSets.items():
            arcrole, ELR, linkqname, arcqname = baseSetKey
            if ELR and linkqname and arcqname and not arcrole.startswith("XBRL-"):
                # assure summationItem, then parentChild, then others
                if not (arcroleFilter == arcrole or
                        arcroleFilter == "*" and arcrole not in (XbrlConst.summationItem, XbrlConst.parentChild)):
                    continue
                if arcrole == XbrlConst.parentChild:
                    ineffectiveArcs = ModelRelationshipSet.ineffectiveArcs(baseSetModelLinks, arcrole)
                    #validate ineffective arcs
                    for modelRel in ineffectiveArcs:
                        if isinstance(modelRel.fromModelObject, ModelObject) and isinstance(modelRel.toModelObject, ModelObject):
                            modelXbrl.error("SBR.NL.2.3.4.06",
                                _("Ineffective arc %(arc)s in \nlink role %(linkrole)s \narcrole %(arcrole)s \nfrom %(conceptFrom)s \nto %(conceptTo)s \n%(ineffectivity)s"),
                                modelObject=modelRel, arc=modelRel.qname, arcrole=modelRel.arcrole,
                                linkrole=modelRel.linkrole, linkroleDefinition=modelXbrl.roleTypeDefinition(modelRel.linkrole),
                                conceptFrom=modelRel.fromModelObject.qname, conceptTo=modelRel.toModelObject.qname,
                                ineffectivity=modelRel.ineffectivity)
                if arcrole == XbrlConst.parentChild:
                    isStatementSheet = any(linkroleDefinitionStatementSheet.match(roleType.definition or '')
                                           for roleType in val.modelXbrl.roleTypes.get(ELR,()))
                    conceptsPresented = set()
                    # 6.12.2 check for distinct order attributes
                    parentChildRels = modelXbrl.relationshipSet(arcrole, ELR)
                    for relFrom, siblingRels in parentChildRels.fromModelObjects().items():
                        targetConceptPreferredLabels = defaultdict(dict)
                        orderRels = {}
                        firstRel = True
                        relFromUsed = True
                        for rel in siblingRels:
                            if firstRel:
                                firstRel = False
                                if relFrom in conceptsUsed:
                                    conceptsUsed[relFrom] = True # 6.12.3, has a pres relationship
                                    relFromUsed = True
                            relTo = rel.toModelObject
                            preferredLabel = rel.preferredLabel
                            if relTo in conceptsUsed:
                                conceptsUsed[relTo] = True # 6.12.3, has a pres relationship
                                if preferredLabel and preferredLabel != "":
                                    conceptRelsUsedWithPreferredLabels[relTo].append(rel)
                                    if preferredLabel in ("periodStart","periodEnd"):
                                        modelXbrl.error("SBR.NL.2.3.4.03",
                                            _("Preferred label on presentation relationships not allowed"), modelObject=modelRel)
                                # 6.12.5 distinct preferred labels in base set
                                preferredLabels = targetConceptPreferredLabels[relTo]
                                if (preferredLabel in preferredLabels or
                                    (not relFrom.isTuple and
                                     (not preferredLabel or None in preferredLabels))):
                                    if preferredLabel in preferredLabels:
                                        rel2, relTo2 = preferredLabels[preferredLabel]
                                    else:
                                        rel2 = relTo2 = None
                                    modelXbrl.error("SBR.NL.2.3.4.06",
                                        _("Concept %(concept)s has duplicate preferred label %(preferredLabel)s in link role %(linkrole)s"),
                                        modelObject=(rel, relTo, rel2, relTo2),
                                        concept=relTo.qname, fromConcept=rel.fromModelObject.qname,
                                        preferredLabel=preferredLabel, linkrole=rel.linkrole, linkroleDefinition=modelXbrl.roleTypeDefinition(rel.linkrole))
                                else:
                                    preferredLabels[preferredLabel] = (rel, relTo)
                                if relFromUsed:
                                    # 6.14.5
                                    conceptsPresented.add(relFrom.objectIndex)
                                    conceptsPresented.add(relTo.objectIndex)
                            order = rel.order
                            if order in orderRels:
                                modelXbrl.error("SBR.NL.2.3.4.05",
                                    _("Duplicate presentation relations from concept %(conceptFrom)s for order %(order)s in base set role %(linkrole)s to concept %(conceptTo)s and to concept %(conceptTo2)s"),
                                    modelObject=(rel, orderRels[order]), conceptFrom=relFrom.qname, order=rel.arcElement.get("order"), linkrole=rel.linkrole, linkroleDefinition=modelXbrl.roleTypeDefinition(rel.linkrole),
                                    conceptTo=rel.toModelObject.qname, conceptTo2=orderRels[order].toModelObject.qname)
                            else:
                                orderRels[order] = rel
                            if not relFrom.isTuple:
                                if relTo in localPreferredLabels:
                                    if {None, preferredLabel} & localPreferredLabels[relTo]:
                                        val.modelXbrl.error("SBR.NL.2.3.4.06",
                                            _("Non-distinguished preferredLabel presentation relations from concept %(conceptFrom)s in base set role %(linkrole)s"),
                                            modelObject=rel, conceptFrom=relFrom.qname, linkrole=rel.linkrole, conceptTo=relTo.qname)
                                localPreferredLabels[relTo].add(preferredLabel)
                        targetConceptPreferredLabels.clear()
                        orderRels.clear()
                    localPreferredLabels.clear() # clear for next relationship
                    for conceptPresented in conceptsPresented:
                        if conceptPresented in usedCalcsPresented:
                            usedCalcPairingsOfConcept = usedCalcsPresented[conceptPresented]
                            if len(usedCalcPairingsOfConcept & conceptsPresented) > 0:
                                usedCalcPairingsOfConcept -= conceptsPresented
                elif arcrole == XbrlConst.summationItem:
                    # find a calc relationship to get the containing document name
                    for modelRel in val.modelXbrl.relationshipSet(arcrole, ELR).modelRelationships:
                        val.modelXbrl.error("SBR.NL.2.3.9.01",
                            _("Calculation linkbase linkrole %(linkrole)s"),
                            modelObject=modelRel, linkrole=ELR)
                        break

                elif arcrole == XbrlConst.all or arcrole == XbrlConst.notAll:
                    drsELRs.add(ELR)

                else:
                    if arcrole == XbrlConst.dimensionDefault:
                        for modelRel in val.modelXbrl.relationshipSet(arcrole).modelRelationships:
                            val.modelXbrl.error("SBR.NL.2.3.6.05",
                                _("Dimension-default in from %(conceptFrom)s to %(conceptTo)s in role %(linkrole)s is not allowed"),
                                modelObject=modelRel, conceptFrom=modelRel.fromModelObject.qname, conceptTo=modelRel.toModelObject.qname,
                                linkrole=modelRel.linkrole)
                    ''' removed per RH 2013-01-11
                    if not (XbrlConst.isStandardArcrole(arcrole) or XbrlConst.isDefinitionOrXdtArcrole(arcrole)):
                        for modelRel in val.modelXbrl.relationshipSet(arcrole).modelRelationships:
                            relTo = modelRel.toModelObject
                            relFrom = modelRel.fromModelObject
                            if not ((isinstance(relFrom,ModelConcept) and isinstance(relTo,ModelConcept)) or
                                    (relFrom.modelDocument.inDTS and
                                     (relTo.qname == XbrlConst.qnGenLabel and modelRel.arcrole == XbrlConst.elementLabel) or
                                     (relTo.qname == XbrlConst.qnGenReference and modelRel.arcrole == XbrlConst.elementReference) or
                                     (relTo.qname == val.qnSbrLinkroleorder))):
                                val.modelXbrl.error("SBR.NL.2.3.2.07",
                                    _("The source and target of an arc must be in the DTS from %(elementFrom)s to %(elementTo)s, in linkrole %(linkrole)s, arcrole %(arcrole)s"),
                                    modelObject=modelRel, elementFrom=relFrom.qname, elementTo=relTo.qname,
                                    linkrole=modelRel.linkrole, arcrole=arcrole)
                        '''

    del localPreferredLabels # dereference
    del usedCalcFromTosELR
    del val.summationItemRelsSetAllELRs

    val.modelXbrl.profileActivity("... filer relationships checks", minTimeToShow=1.0)


    # checks on dimensions
    checkFilingDimensions(val, drsELRs)
    val.modelXbrl.profileActivity("... filer dimensions checks", minTimeToShow=1.0)

    del conceptRelsUsedWithPreferredLabels

    # 6 16 4, 1.16.5 Base sets of Domain Relationship Sets testing
    val.modelXbrl.profileActivity("... filer preferred label checks", minTimeToShow=1.0)

    # moved from original validateSBRnl finally

    for qname, modelType in modelXbrl.qnameTypes.items():
        if qname.namespaceURI not in val.disclosureSystem.baseTaxonomyNamespaces:
            facets = modelType.facets
            if facets:
                lengthFacets = facets.keys() & {"minLength", "maxLength", "length"}
                if lengthFacets:
                    modelXbrl.error("SBR.NL.2.2.7.02",
                        _("Type %(typename)s has length restriction facets %(facets)s"),
                        modelObject=modelType, typename=modelType.qname, facets=", ".join(lengthFacets))
                if "enumeration" in facets and not modelType.isDerivedFrom(XbrlConst.qnXbrliStringItemType):
                    modelXbrl.error("SBR.NL.2.2.7.04",
                        _("Concept %(concept)s has enumeration and is not based on stringItemType"),
                        modelObject=modelType, concept=modelType.qname)

    ''' removed RH 2011-12-23, corresponding use of nameWordsTable in ValidateFilingDTS
    # build camelCasedNamesTable
    self.nameWordsTable = {}
    for name in modelXbrl.nameConcepts.keys():
        words = []
        wordChars = []
        lastchar = ""
        for c in name:
            if c.isupper() and lastchar.islower(): # it's another word
                partialName = ''.join(wordChars)
                if partialName in modelXbrl.nameConcepts:
                    words.append(partialName)
            wordChars.append(c)
            lastchar = c
        if words:
            self.nameWordsTable[name] = words
    self.modelXbrl.profileActivity("... build name words table", minTimeToShow=1.0)
    '''



    # check presentation link roles for generic linkbase order number
    ordersRelationshipSet = modelXbrl.relationshipSet("http://www.nltaxonomie.nl/2011/arcrole/linkrole-order")
    presLinkroleNumberURI = {}
    presLinkrolesCount = 0
    for countLinkroles in (True, False):
        for _roleURI, modelRoleTypes in modelXbrl.roleTypes.items():
            for modelRoleType in modelRoleTypes:
                if XbrlConst.qnLinkPresentationLink in modelRoleType.usedOns:
                    if countLinkroles:
                        presLinkrolesCount += 1
                    else:
                        if not ordersRelationshipSet:
                            modelXbrl.error("SBR.NL.2.2.3.06",
                                _("Presentation linkrole %(linkrole)s missing order number relationship set"),
                                modelObject=modelRoleType, linkrole=modelRoleType.roleURI)
                        else:
                            order = None
                            for orderNumRel in ordersRelationshipSet.fromModelObject(modelRoleType):
                                order = getattr(orderNumRel.toModelObject, "xValue", "(noPSVIvalue)")
                                if order in presLinkroleNumberURI:
                                    modelXbrl.error("SBR.NL.2.2.3.06",
                                        _("Presentation linkrole order number %(order)s of %(linkrole)s also used in %(otherLinkrole)s"),
                                        modelObject=modelRoleType, order=order, linkrole=modelRoleType.roleURI, otherLinkrole=presLinkroleNumberURI[order])
                                else:
                                    presLinkroleNumberURI[order] = modelRoleType.roleURI
                            if not order:
                                modelXbrl.error("SBR.NL.2.2.3.06",
                                    _("Presentation linkrole %(linkrole)s missing order number"),
                                    modelObject=modelRoleType, linkrole=modelRoleType.roleURI)
        if countLinkroles and presLinkrolesCount < 2:
            break   # don't check order numbers if only one presentation linkrole
    # check arc role definitions for labels
    for arcroleURI, modelRoleTypes in modelXbrl.arcroleTypes.items():
        for modelRoleType in modelRoleTypes:
            if (not arcroleURI.startswith("http://xbrl.org/") and
                modelRoleType.modelDocument.targetNamespace not in val.disclosureSystem.baseTaxonomyNamespaces and
                (not modelRoleType.genLabel(lang="nl") or not modelRoleType.genLabel(lang="en"))):
                modelXbrl.error("SBR.NL.2.2.4.02",
                    _("ArcroleType missing nl or en generic label: %(arcrole)s"),
                    modelObject=modelRoleType, arcrole=arcroleURI)

    for domainElt in val.typedDomainElements:
        if domainElt.modelDocument.targetNamespace not in val.disclosureSystem.baseTaxonomyNamespaces:
            if not domainElt.genLabel(fallbackToQname=False, lang="nl"):
                modelXbrl.error("SBR.NL.2.2.8.01",
                    _("Typed dimension domain element %(concept)s must have a generic label"),
                    modelObject=domainElt, concept=domainElt.qname)
            if domainElt.type is not None and domainElt.type.localName == "complexType":
                modelXbrl.error("SBR.NL.2.2.8.02",
                    _("Typed dimension domain element %(concept)s has disallowed complex content"),
                    modelObject=domainElt, concept=domainElt.qname)

    modelXbrl.profileActivity("... SBR role types and type facits checks", minTimeToShow=1.0)
    # end moved from ValidateFiling

    # 3.2.4.4 check each using prefix against taxonomy declaring the prefix
    for docs in modelXbrl.namespaceDocs.values():
        for doc in docs:
            for prefix, NS in doc.xmlRootElement.nsmap.items():
                if NS in val.namespacePrefix and prefix != val.namespacePrefix[NS]:
                    modelXbrl.error("SBR.NL.3.2.4.04",
                        _("The assigned namespace prefix %(assignedPrefix)s for the schema that declares the targetnamespace %(namespace)s, MUST be adhired by all other NT schemas, referencedPrefix: %(referencedPrefix)s"),
                        modelObject=doc.xmlRootElement, namespace=NS, assignedPrefix=val.namespacePrefix.get(NS, ''), referencedPrefix=prefix)

    # check non-concept elements that can appear in elements for labels (concepts checked by
    labelsRelationshipSet = modelXbrl.relationshipSet((XbrlConst.conceptLabel, XbrlConst.elementLabel))
    baseTaxonomyNamespaces = val.disclosureSystem.baseTaxonomyNamespaces
    for eltDef in modelXbrl.qnameConcepts.values():
        if (not (eltDef.isItem or eltDef.isTuple or eltDef.isLinkPart) and
            eltDef.qname.namespaceURI not in baseTaxonomyNamespaces):
            eltDefHasDefaultLangStandardLabel = False
            for modelLabelRel in labelsRelationshipSet.fromModelObject(eltDef):
                modelLabel = modelLabelRel.toModelObject
                role = modelLabel.role
                text = modelLabel.text
                lang = modelLabel.xmlLang
                if text and lang and val.disclosureSystem.defaultXmlLang and lang.startswith(val.disclosureSystem.defaultXmlLang):
                    if role in (XbrlConst.standardLabel, XbrlConst.genStandardLabel):
                        eltDefHasDefaultLangStandardLabel = True
            if not eltDefHasDefaultLangStandardLabel:
                modelXbrl.error("SBR.NL.3.2.15.01",
                    _("XML nodes that can appear in instances MUST have standard labels in the local language: %(element)s"),
                    modelObject=eltDef, element=eltDef.qname)

    val.modelXbrl.profileStat(_("validate{0}").format(modelXbrl.modelManager.disclosureSystem.validationType))

    modelXbrl.modelManager.showStatus(_("ready"), 2000)

def isStandardUri(val, uri):
    try:
        return val._isStandardUri[uri]
    except KeyError:
        isStd = (uri in val.disclosureSystem.standardTaxonomiesDict or
                 (not isHttpUrl(uri) and
                  # try 2011-12-23 RH: if works, remove the localHrefs
                  # any(u.endswith(e) for u in (uri.replace("\\","/"),) for e in disclosureSystem.standardLocalHrefs)
                  "/basis/sbr/" in uri.replace("\\","/")
                  ))
        val._isStandardUri[uri] = isStd
        return isStd

def checkConceptLabels(val, modelXbrl, labelsRelationshipSet, disclosureSystem, concept):
    hasDefaultLangStandardLabel = False
    dupLabels = {}
    for modelLabelRel in labelsRelationshipSet.fromModelObject(concept):
        modelLabel = modelLabelRel.toModelObject
        if isinstance(modelLabel, ModelResource) and modelLabel.xmlLang:
            if modelLabel.xmlLang.startswith(disclosureSystem.defaultXmlLang) and \
               modelLabel.role == XbrlConst.standardLabel:
                hasDefaultLangStandardLabel = True
            dupDetectKey = ( (modelLabel.role or ''), modelLabel.xmlLang)
            if dupDetectKey in dupLabels:
                modelXbrl.error("SBR.NL.2.2.1.05",
                    _("Concept %(concept)s has duplicated labels for role %(role)s lang %(lang)s."),
                    modelObject=(modelLabel, dupLabels[dupDetectKey]), # removed concept from modelObjects
                    concept=concept.qname, role=dupDetectKey[0], lang=dupDetectKey[1])
            else:
                dupLabels[dupDetectKey] = modelLabel
            if modelLabel.role in (XbrlConst.periodStartLabel, XbrlConst.periodEndLabel):
                modelXbrl.error("SBR.NL.2.3.8.03",
                    _("Concept %(concept)s has label for semantical role %(role)s."),
                    modelObject=modelLabel, concept=concept.qname, role=modelLabel.role)
    for role, lang in dupLabels.keys():
        if role and lang != disclosureSystem.defaultXmlLang and (role,disclosureSystem.defaultXmlLang) not in dupLabels:
            modelXbrl.error("SBR.NL.2.3.8.05",
                _("Concept %(concept)s has en but no nl label in role %(role)s."),
                modelObject=(concept,dupLabels[(role,lang)]), concept=concept.qname, role=role)



# for SBR 2.3.4.01
def pLinkedNonAbstractDescendantQnames(modelXbrl, concept, descendants=None):
    if descendants is None: descendants = set()
    for rel in modelXbrl.relationshipSet(XbrlConst.parentChild).fromModelObject(concept):
        child = rel.toModelObject
        if child is not None:
            if child.isAbstract:
                pLinkedNonAbstractDescendantQnames(modelXbrl, child, descendants)
            else:
                descendants.add(child.qname)
    return descendants



from __future__ import annotations

from tests.integration_tests.validation.conformance_suite_config import ConformanceSuiteConfig
from tests.integration_tests.validation.conformance_suite_configurations.efm_current import config as efm_current
from tests.integration_tests.validation.conformance_suite_configurations.esef_ixbrl_2021 import config as esef_ixbrl_2021
from tests.integration_tests.validation.conformance_suite_configurations.esef_ixbrl_2022 import config as esef_ixbrl_2022
from tests.integration_tests.validation.conformance_suite_configurations.esef_xhtml_2021 import config as esef_xhtml_2021
from tests.integration_tests.validation.conformance_suite_configurations.esef_xhtml_2022 import config as esef_xhtml_2022
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_2_1 import config as xbrl_2_1
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_calculations_1_1 import config as xbrl_calculations_1_1
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_dimensions_1_0 import config as xbrl_dimensions_1_0
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_extensible_enumerations_1_0 import config as xbrl_extensible_enumerations_1_0
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_extensible_enumerations_2_0 import config as xbrl_extensible_enumerations_2_0
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_formula_1_0 import config as xbrl_formula_1_0
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_formula_1_0_assertion_severity_2_0 import config as xbrl_formula_1_0_assertion_severity_2_0
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_formula_1_0_function_registry import config as xbrl_formula_1_0_function_registry
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_ixbrl_1_1 import config as xbrl_ixbrl_1_1
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_link_role_registry_1_0 import config as xbrl_link_role_registry_1_0
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_oim_1_0 import config as xbrl_oim_1_0
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_table_linkbase_1_0 import config as xbrl_table_linkbase_1_0
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_taxonomy_packages_1_0 import config as xbrl_taxonomy_packages_1_0
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_transformation_registry_3 import config as xbrl_transformation_registry_3
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_transformation_registry_4 import config as xbrl_transformation_registry_4
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_transformation_registry_5 import config as xbrl_transformation_registry_5
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_utr_malformed_1_0 import configs as xbrl_utr_malformed_1_0
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_utr_registry_1_0 import config as xbrl_utr_registry_1_0
from tests.integration_tests.validation.conformance_suite_configurations.xbrl_utr_structure_1_0 import config as xbrl_utr_structure_1_0


ALL_CONFORMANCE_SUITE_CONFIGS: tuple[ConformanceSuiteConfig, ...] = (
    efm_current,
    esef_ixbrl_2021,
    esef_ixbrl_2022,
    esef_xhtml_2021,
    esef_xhtml_2022,
    xbrl_2_1,
    xbrl_calculations_1_1,
    xbrl_dimensions_1_0,
    xbrl_extensible_enumerations_1_0,
    xbrl_extensible_enumerations_2_0,
    xbrl_formula_1_0,
    xbrl_formula_1_0_assertion_severity_2_0,
    xbrl_formula_1_0_function_registry,
    xbrl_ixbrl_1_1,
    xbrl_link_role_registry_1_0,
    xbrl_oim_1_0,
    xbrl_table_linkbase_1_0,
    xbrl_taxonomy_packages_1_0,
    xbrl_transformation_registry_3,
    xbrl_transformation_registry_4,
    xbrl_transformation_registry_5,
    *xbrl_utr_malformed_1_0,
    xbrl_utr_registry_1_0,
    xbrl_utr_structure_1_0,
)

PUBLIC_CONFORMANCE_SUITE_CONFIGS = tuple(c for c in ALL_CONFORMANCE_SUITE_CONFIGS if c.public_download_url)



'''
See COPYRIGHT.md for copyright information.
'''
from __future__ import annotations

import datetime
import regex as re
from arelle import ModelValue
from arelle.formula import XPathContext
from arelle.FunctionUtil import (anytypeArg, atomicArg, stringArg, numericArg, qnameArg, nodeArg)
from arelle.XmlValidate import lexicalPatterns
from arelle.formula.XPathParser import OperationDef, ProgHeader
from math import isnan, fabs, isinf
from decimal import Decimal, InvalidOperation
from numbers import Number

class FORG0001(Exception):
    def __init__(self, message=None):
        self.message = message
        self.args = ( self.__repr__(), )
    def __repr__(self):
        return _("Exception: FORG0001, invalid constructor")

class FONS0004(Exception):
    def __init__(self, message=None):
        self.message = message
        self.args = ( self.__repr__(), )
    def __repr__(self):
        return _("Exception: FONS0004, no namespace found for prefix")

class xsFunctionNotAvailable(Exception):
    def __init__(self):
        self.args =  (_("xs function not available"),)
    def __repr__(self):
        return self.args[0]

def call(
        xc: XPathContext.XPathContext,
        p: ProgHeader | OperationDef | None,
        localname: str,
        args: XPathContext.ResultStack,
) -> XPathContext.RecursiveContextItem:
    source = atomicArg(xc, p, args, 0, "value?", missingArgFallback=() )
    if source == (): return source
    try:
        if localname not in xsFunctions: raise xsFunctionNotAvailable
        return xsFunctions[localname](xc, p, source)
    except (FORG0001, ValueError, TypeError) as ex:
        if hasattr(ex, "message") and ex.message:
            exMsg = ", " + ex.message
        else:
            exMsg = ""
        raise XPathContext.XPathException(p, 'err:FORG0001',
                                          _('invalid cast from {0} to xs:{1}{2}').format(
                                            type(source).__name__,
                                            localname,
                                            exMsg))
    except xsFunctionNotAvailable:
        raise XPathContext.FunctionNotAvailable("xs:{0}".format(localname))

objtype = {
        #'untypedAtomic': untypedAtomic,
        'dateTime':  ModelValue.DateTime,
        'date': ModelValue.DateTime,
        'time': ModelValue.Time,
        #'duration': duration,
        'yearMonthDuration': ModelValue.YearMonthDuration,
        'dayTimeDuration': ModelValue.DayTimeDuration,
        'float': float,
        'double': float,
        'decimal': Decimal,
        'integer': int,
        'nonPositiveInteger': int,
        'negativeInteger': int,
        'long': int,
        'int': int,
        'short': int,
        'byte': int,
        'nonNegativeInteger': int,
        'unsignedLong': int,
        'unsignedInt': int,
        'unsignedShort': int,
        'unsignedByte': int,
        'positiveInteger': int,
        #'gYearMonth': gYearMonth,
        #'gYear': gYear,
        #'gMonthDay': gMonthDay,
        #'gDay': gDay,
        #'gMonth': gMonth,
        'string': str,
        'normalizedString': str,
        'token': str,
        'language': str,
        'NMTOKEN': str,
        'Name': str,
        'NCName': str,
        'ID': str,
        'IDREF': str,
        'ENTITY': str,
        'boolean': bool,
        #'base64Binary': byte,
        #'hexBinary': byte,
        'anyURI': ModelValue.AnyURI,
        'QName': ModelValue.QName,
        'NOTATION': str,
      }

def isXsType(localName):
    if localName[-1] in ('?', '+', '*'):
        return localName[:-1] in xsFunctions
    return localName in xsFunctions

def untypedAtomic(xc, p, source):
    raise xsFunctionNotAvailable()

def anyType(xc, p, source):
    raise xsFunctionNotAvailable()

def anyAtomicType(xc, p, source):
    raise xsFunctionNotAvailable()

def dateTime(xc, p, source):
    if isinstance(source,datetime.datetime) and source.dateOnly is False: return source
    return ModelValue.dateTime(source, type=ModelValue.DATETIME, castException=FORG0001)

def dateTimeInstantEnd(xc, p, source):
    if isinstance(source,datetime.datetime): return source  # true for either datetime.date or datetime.datetime
    return ModelValue.dateTime(source, addOneDay=True, type=ModelValue.DATETIME, castException=FORG0001)

def xbrliDateUnion(xc, p, source):
    if isinstance(source,datetime.date): return source  # true for either datetime.date or datetime.datetime
    return ModelValue.dateTime(source, type=ModelValue.DATEUNION, castException=FORG0001)

def date(xc, p, source):
    return ModelValue.dateTime(source, type=ModelValue.DATE, castException=FORG0001)

def time(xc, p, source):
    return ModelValue.time(source, castException=FORG0001)

def duration(xc, p, source):
    raise xsFunctionNotAvailable()

def yearMonthDuration(xc, p, source):
    return ModelValue.yearMonthDuration(source)

def dayTimeDuration(xc, p, source):
    return ModelValue.dayTimeDuration(source)

def xs_float(xc, p, source):
    try:
        return float(source)
    except ValueError:
        raise FORG0001

def double(xc, p, source):
    try:
        return float(source)
    except ValueError:
        raise FORG0001

def decimal(xc, p, source):
    try:
        return Decimal(source)
    except InvalidOperation:
        raise FORG0001

def integer(xc, p, source):
    try:
        return int(source)
    except ValueError:
        raise FORG0001

def nonPositiveInteger(xc, p, source):
    try:
        i = int(source)
        if i <= 0: return i
    except ValueError:
        pass
    raise FORG0001

def negativeInteger(xc, p, source):
    try:
        i = int(source)
        if i < 0: return i
    except ValueError:
        pass
    raise FORG0001

def long(xc, p, source):
    try:
        return int(source)
    except ValueError:
        raise FORG0001

def xs_int(xc, p, source):
    try:
        i = int(source)
        if i <= 2147483647 and i >= -2147483648: return i
    except ValueError:
        pass
    raise FORG0001

def short(xc, p, source):
    try:
        i = int(source)
        if i <= 32767 and i >= -32767: return i
    except ValueError:
        pass
    raise FORG0001

def byte(xc, p, source):
    try:
        i = int(source)
        if i <= 127 and i >= -128: return i
    except ValueError:
        pass
    raise FORG0001

def nonNegativeInteger(xc, p, source):
    try:
        i = int(source)
        if i >= 0: return i
    except ValueError:
        pass
    raise FORG0001

def unsignedLong(xc, p, source):
    try:
        i = int(source)
        if i >= 0: return i
    except ValueError:
        pass
    raise FORG0001

def unsignedInt(xc, p, source):
    try:
        i = int(source)
        if i <= 4294967295 and i >= 0: return i
    except ValueError:
        pass
    raise FORG0001

def unsignedShort(xc, p, source):
    try:
        i = int(source)
        if i <= 65535 and i >= 0: return i
    except ValueError:
        pass
    raise FORG0001

def unsignedByte(xc, p, source):
    try:
        i = int(source)
        if i <= 255 and i >= 0: return i
    except ValueError:
        pass
    raise FORG0001

def positiveInteger(xc, p, source):
    try:
        i = int(source)
        if i > 0: return i
    except ValueError:
        pass
    raise FORG0001

def gYearMonth(xc, p, source):
    try:
        match = lexicalPatterns['gYearMonth'].match(source)
        if match:
            year, month, zSign, zHrMin, zHr, zMin = match.groups()
            return ModelValue.gYearMonth(year, month)
    except (ValueError, TypeError):
        pass
    raise FORG0001

def gYear(xc, p, source):
    try:
        match = lexicalPatterns['gYear'].match(source)
        if match:
            year, zSign, zHrMin, zHr, zMin = match.groups()
            return ModelValue.gYear(year)
    except (ValueError, TypeError):
        pass
    raise FORG0001

def gMonthDay(xc, p, source):
    try:
        match = lexicalPatterns['gMonthDay'].match(source)
        if match:
            month, day, zSign, zHrMin, zHr, zMin = match.groups()
            if not int(day) > {2:29, 4:30, 6:30, 9:30, 11:30, 1:31, 3:31, 5:31, 7:31, 8:31, 10:31, 12:31}[int(month)]:
                return ModelValue.gMonthDay(month, day)
    except (ValueError, TypeError):
        pass
    raise FORG0001

def gDay(xc, p, source):
    try:
        match = lexicalPatterns['gDay'].match(source)
        if match:
            day, zSign, zHrMin, zHr, zMin = match.groups()
            return ModelValue.gDay(day)
    except (ValueError, TypeError):
        pass
    raise FORG0001

def gMonth(xc, p, source):
    try:
        match = lexicalPatterns['gMonth'].match(source)
        if match:
            month, zSign, zHrMin, zHr, zMin = match.groups()
            return ModelValue.gMonth(month)
    except (ValueError, TypeError):
        pass
    raise FORG0001

def xsString(xc, p, source):
    if isinstance(source,bool):
        return 'true' if source else 'false'
    elif isinstance(source,float):
        if isnan(source):
            return "NaN" # note that -NaN is reported as NaN in XML
        elif isinf(source):
            return "-INF" if source < 0 else "INF"
        '''
        numMagnitude = fabs(source)
        if numMagnitude < 1000000 and numMagnitude > .000001:
            # don't want floating notation which python does for more than 4 decimal places
            s =
        '''
        s = str(source)
        if s.endswith(".0"):
            s = s[:-2]
        return s
    elif isinstance(source,Decimal):
        if isnan(source):
            return "NaN" # note that -NaN is reported as NaN in XML
        elif isinf(source):
            return "-INF" if source < 0 else "INF"
        return "{:f}".format(source) # str(source) causes unwanted scientific notation for values like 0.00000001
    elif isinstance(source,ModelValue.DateTime):
        return ('{0:%Y-%m-%d}' if source.dateOnly else '{0:%Y-%m-%dT%H:%M:%S}').format(source)
    return str(source)

def normalizedString(xc, p, source):
    return str(source)

tokenPattern = re.compile(r"^\s*([-\.:\w]+)\s*$")
def token(xc, p, source):
    s = str(source)
    if tokenPattern.match(s): return s
    raise FORG0001

languagePattern = re.compile("[a-zA-Z]{1,8}(-[a-zA-Z0-9]{1,8})*")
def language(xc, p, source):
    s = str(source)
    if languagePattern.match(s): return s
    raise FORG0001

def NMTOKEN(xc, p, source):
    raise xsFunctionNotAvailable()

def Name(xc, p, source):
    raise xsFunctionNotAvailable()

def NCName(xc, p, source):
    raise xsFunctionNotAvailable()

def ID(xc, p, source):
    raise xsFunctionNotAvailable()

def IDREF(xc, p, source):
    raise xsFunctionNotAvailable()

def ENTITY(xc, p, source):
    raise xsFunctionNotAvailable()

def boolean(xc, p, source):
    if isinstance(source,bool):
        return source
    elif isinstance(source, Number):
        if source == 1:
            return True
        elif source == 0:
            return False
    elif isinstance(source,str):
        b = source.lower()
        if b in ('true','yes'):
            return True
        elif b in ('false','no'):
            return False
    raise FORG0001

def base64Binary(xc, p, source):
    raise xsFunctionNotAvailable()

def hexBinary(xc, p, source):
    raise xsFunctionNotAvailable()

def anyURI(xc, p, source):
    return ModelValue.anyURI(source)

def QName(xc, p, source):
    if isinstance(p, ProgHeader):
        element = p.element
    elif xc.progHeader:
        element = xc.progHeader.element
    else:
        element = xc.sourceElement
    return ModelValue.qname(element, source, castException=FORG0001, prefixException=FONS0004)

def NOTATION(xc, p, source):
    raise xsFunctionNotAvailable()

xsFunctions = {
    'untypedAtomic': untypedAtomic,
    'anyType': anyType,
    'anyAtomicType': anyAtomicType,
    'dateTime': dateTime,
    'DATETIME_START': dateTime,
    'DATETIME_INSTANT_END': dateTimeInstantEnd,
    'XBRLI_DATEUNION': xbrliDateUnion,
    'date': date,
    'time': time,
    'duration': duration,
    'yearMonthDuration': yearMonthDuration,
    'dayTimeDuration': dayTimeDuration,
    'float': xs_float,
    'double': double,
    'decimal': decimal,
    'integer': integer,
    'nonPositiveInteger': nonPositiveInteger,
    'negativeInteger': negativeInteger,
    'long': long,
    'int': xs_int,
    'short': short,
    'byte': byte,
    'nonNegativeInteger': nonNegativeInteger,
    'unsignedLong': unsignedLong,
    'unsignedInt': unsignedInt,
    'unsignedShort': unsignedShort,
    'unsignedByte': unsignedByte,
    'positiveInteger': positiveInteger,
    'gYearMonth': gYearMonth,
    'gYear': gYear,
    'gMonthDay': gMonthDay,
    'gDay': gDay,
    'gMonth': gMonth,
    'string': xsString,
    'normalizedString': normalizedString,
    'token': token,
    'language': language,
    'NMTOKEN': NMTOKEN,
    'Name': Name,
    'NCName': NCName,
    'ID': ID,
    'IDREF': IDREF,
    'ENTITY': ENTITY,
    'boolean': boolean,
    'base64Binary': base64Binary,
    'hexBinary': hexBinary,
    'anyURI': anyURI,
    'QName': QName,
    'NOTATION': NOTATION,
    }



from __future__ import annotations

import sys
from argparse import ArgumentParser, Namespace
from tests.integration_tests.validation.validation_util import get_conformance_suite_test_results
from tests.integration_tests.validation.conformance_suite_config import ConformanceSuiteConfig
from tests.integration_tests.validation.conformance_suite_configs import (
    ALL_CONFORMANCE_SUITE_CONFIGS,
    PUBLIC_CONFORMANCE_SUITE_CONFIGS
)
from tests.integration_tests.validation.download_conformance_suites import (
    download_conformance_suite, extract_conformance_suite
)
from typing import Any, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from _pytest.mark import ParameterSet


ARGUMENTS: list[dict[str, Any]] = [
    {
        "name": "--all",
        "action": "store_true",
        "help": "Select all configured conformance suites"
    },
    {
        "name": "--download-overwrite",
        "action": "store_true",
        "help": "Download (and overwrite) selected conformance suite files"
    },
    {
        "name": "--download-missing",
        "action": "store_true",
        "help": "Download missing selected conformance suite files"
    },
    {
        "name": "--list",
        "action": "store_true",
        "help": "List names of all configured conformance suites"
    },
    {
        "name": "--log-to-file",
        "action": "store_true",
        "help": "Writes logs and results to .txt and .csv files"
    },
    {
        "name": "--name",
        "action": "store",
        "help": "Select only conformance suites with given names, comma delimited"
    },
    {
        "name": "--offline",
        "action": "store_true",
        "help": "Run without loading anything from the internet (local files and cache only)"
    },
    {
        "name": "--public",
        "action": "store_true",
        "help": "Select all public conformance suites"
    },
    {
        "name": "--shard",
        "action": "store",
        "help": "0-indexed shard to run",
        "type": int,
    },
    {
        "name": "--test",
        "action": "store_true",
        "help": "Run selected conformance suite tests"
    },
]
DOWNLOAD_MISSING = 'missing'
DOWNLOAD_OVERWRITE = 'overwrite'
SELECT_ALL = 'all'
SELECT_PUBLIC = 'public'


def _get_conformance_suite_names(select_option: str) -> tuple[ConformanceSuiteConfig, ...]:
    if select_option == SELECT_ALL:
        return ALL_CONFORMANCE_SUITE_CONFIGS
    elif select_option == SELECT_PUBLIC:
        return PUBLIC_CONFORMANCE_SUITE_CONFIGS
    elif select_option:
        filter_list = select_option.split(',')
        names = []
        for filter_item in filter_list:
            match_configs = [c for c in ALL_CONFORMANCE_SUITE_CONFIGS if c.name == filter_item]
            if not match_configs:
                raise ValueError(f'Provided name "{filter_item}" did not match any configured conformance suite names.')
            names.extend(match_configs)
        return tuple(names)
    else:
        raise ValueError('Please use --all, --public, or --name to specify which conformance suites to use.')


def run_conformance_suites(
        select_option: str,
        test_option: bool,
        shard: int | None,
        download_option: str | None = None,
        log_to_file: bool = False,
        offline_option: bool = False) -> list[ParameterSet]:
    conformance_suite_configs = _get_conformance_suite_names(select_option)
    if download_option:
        overwrite = download_option == DOWNLOAD_OVERWRITE
        for conformance_suite_config in conformance_suite_configs:
            download_conformance_suite(conformance_suite_config, overwrite=overwrite)
    for conformance_suite_config in conformance_suite_configs:
        extract_conformance_suite(conformance_suite_config)
    all_results = []
    if test_option:
        for config in conformance_suite_configs:
            results = get_conformance_suite_test_results(config, shard=shard, log_to_file=log_to_file, offline=offline_option)
            all_results.extend(results)
    return all_results


def run_conformance_suites_options(options: Namespace) -> list[ParameterSet]:
    select_option = get_select_option(options)
    download_option = get_download_option(options)
    assert download_option or options.test, \
        'Specify at least one of download, list, or test.'
    return run_conformance_suites(
        select_option=select_option,
        test_option=options.test,
        shard=options.shard,
        download_option=download_option,
        log_to_file=options.log_to_file,
        offline_option=options.offline
    )


def get_download_option(options: Namespace) -> Optional[str]:
    if options.download_overwrite:
        return DOWNLOAD_OVERWRITE
    elif options.download_missing:
        return DOWNLOAD_MISSING
    return None


def get_select_option(options: Namespace) -> str:
    if options.all:
        return SELECT_ALL
    elif options.public:
        return SELECT_PUBLIC
    assert isinstance(options.name, str)
    return options.name


def run() -> None:
    parser = ArgumentParser(prog=sys.argv[0])
    for arg in ARGUMENTS:
        arg_without_name = {k: v for k, v in arg.items() if k != "name"}
        parser.add_argument(arg["name"], **arg_without_name)
    options = parser.parse_args(sys.argv[1:])
    if options.list:
        for config in ALL_CONFORMANCE_SUITE_CONFIGS:
            print(f'{config.name}\n'
                  f'\tInfo:       {config.info_url}\n'
                  f'\tDownload:   {config.public_download_url or config.membership_url}\n'
                  f'\tLocal Path: {config.prefixed_local_filepath}')
    else:
        run_conformance_suites_options(options)


if __name__ == "__main__":
    run()



"""
See COPYRIGHT.md for copyright information.

This module emits the version.py file contents which are used in the
build process to indicate the time that this version was built.

"""
import datetime
import sys

if __name__ == "__main__":
    timestamp = datetime.datetime.utcnow()
    date_dash_ymd = timestamp.strftime("%Y-%m-%d")

    # add name suffix, like ER3 or TKTABLE
    if len(sys.argv) > 1 and sys.argv[1] and sys.platform not in ("linux",):
        date_dash_ymd += "-" + sys.argv[1]

    if sys.platform == "darwin":
        with open("buildRenameDmg.sh", "w") as fh:
            fh.write("cp dist_dmg/arelle.dmg dist/arelle-macOS-{}.dmg\n".format(date_dash_ymd))
    elif sys.platform.startswith("linux"):
        if len(sys.argv) > 1 and sys.argv[1]:
            sysName = sys.argv[1]
        else:
            sysName = "linux"
        with open("buildRenameLinux-x86_64.sh", "w") as fh:
            fh.write("mv dist/exe.linux-x86_64-{}.{}.tgz dist/arelle-{}-x86_64-{}.tgz\n"
                     .format(sys.version_info[0], sys.version_info[1], sysName, date_dash_ymd))
    elif sys.platform.startswith("win"):
        renameCmdFile = "buildRenamer.bat"
        with open("buildRenameX86.bat", "w") as fh:
            fh.write("rename dist\\arelle-win-x86.exe arelle-win-x86-{}.exe\n".format(date_dash_ymd))
        with open("buildRenameX64.bat", "w") as fh:
            fh.write("rename dist\\arelle-win-x64.exe arelle-win-x64-{0}.exe\n"
                     "rename dist\\arelle-win-x64.zip arelle-win-x64-{0}.zip\n"
                     .format(date_dash_ymd))
        with open("buildRenameSvr27.bat", "w") as fh:
            fh.write("rename dist\\arelle-svr-2.7.zip arelle-svr-2.7-{}.zip\n".format(date_dash_ymd))
        with open("buildRenameZip32.bat", "w") as fh:
            fh.write("rename dist\\arelle-cmd32.zip arelle-cmd32-{}.zip\n".format(date_dash_ymd))
        with open("buildRenameZip64.bat", "w") as fh:
            fh.write("rename dist\\arelle-cmd64.zip arelle-cmd64-{}.zip\n".format(date_dash_ymd))



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from typing import Any

from arelle.ModelDocument import LoadingException, ModelDocument
from arelle.ModelXbrl import ModelXbrl
from arelle.typing import TypeGetText
from arelle.utils.validate.ValidationPlugin import ValidationPlugin
from .PluginValidationDataExtension import PluginValidationDataExtension

_: TypeGetText


class ValidationPluginExtension(ValidationPlugin):
    def newPluginData(self) -> PluginValidationDataExtension:
        return PluginValidationDataExtension(self.name)

    def modelDocumentPullLoader(
        self,
        modelXbrl: ModelXbrl,
        normalizedUri: str,
        filepath: str,
        isEntry: bool,
        namespace: str | None,
        *args: Any,
        **kwargs: Any,
    ) -> ModelDocument | LoadingException | None:
        if self.disclosureSystemFromPluginSelected(modelXbrl):
            return LoadingException(_("XYZ validation plugin is a template for new validation plugins and shouldn't be used directly."))
        return None

    def modelXbrlLoadComplete(
        self,
        modelXbrl: ModelXbrl,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        if self.disclosureSystemFromPluginSelected(modelXbrl):
            if modelXbrl.modelDocument is None:
                modelXbrl.error(
                    codes="XYZ.01.01",
                    msg=_("An XBRL Report Package is required but could not be loaded"),
                    modelObject=modelXbrl,
                )
        return None
